#!/usr/bin/python3

# Copyright (c) 2017-2023 California Institute of Technology ("Caltech"). U.S.
# Government sponsorship acknowledged. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0

'''General utility functions used throughout mrcal

All functions are exported into the mrcal module. So you can call these via
mrcal.utils.fff() or mrcal.fff(). The latter is preferred.

'''

import numpy as np
import numpysane as nps
import sys
import re

import mrcal


def _align_procrustes_points_Rt01_python(p0, p1, weights=None):
    r'''Reference implementation of align_procrustes_points_Rt01() in python

The main implementation is written in C in poseutils.c:

  mrcal_align_procrustes_points_Rt01()

The two implementations are identical, with a test to verify this

    '''
    if weights is None:
        weights = np.ones(p0.shape[:-1], dtype=float)
    return _align_procrustes_points_Rt01(p0,p1,weights)


@nps.broadcast_define( (('N',3,), ('N',3,), ('N',)),
                       (4,3), )
def _align_procrustes_points_Rt01(p0, p1, weights):

    p0 = nps.transpose(p0)
    p1 = nps.transpose(p1)

    # I process Mt instead of M to not need to transpose anything later, and to
    # end up with contiguous-memory results
    Mt = nps.matmult(              (p0 - np.mean(p0, axis=-1)[..., np.newaxis])*weights,
                      nps.transpose(p1 - np.mean(p1, axis=-1)[..., np.newaxis]))
    V,S,Ut = np.linalg.svd(Mt)

    # I look at the second-lowest singular value. One 0 singular value is OK;
    # two isn't
    if S[-2] < 1e-12:
        # Poorly-defined problem. Return error
        return np.zeros((4,3), dtype=float)

    R = nps.matmult(V, Ut)

    # det(R) is now +1 or -1. If it's -1, then this contains a mirror, and thus
    # is not a physical rotation. I compensate by negating the least-important
    # pair of singular vectors
    if np.linalg.det(R) < 0:
        V[:,2] *= -1
        R = nps.matmult(V, Ut)

    # Now that I have my optimal R, I compute the optimal t. From before:
    #
    #   t = mean(a) - R mean(b)
    t = np.mean(p0, axis=-1)[..., np.newaxis] - nps.matmult( R, np.mean(p1, axis=-1)[..., np.newaxis] )

    return nps.glue( R, t.ravel(), axis=-2)


def _align_procrustes_vectors_R01_python(v0, v1, weights=None):
    r'''Reference implementation of align_procrustes_vectors_R01() in python

The main implementation is written in C in poseutils.c:

  mrcal_align_procrustes_vectors_R01()

The two implementations are identical, with a test to verify this

    '''
    if weights is None:
        weights = np.ones(v0.shape[:-1], dtype=float)
    return _align_procrustes_vectors_R01(v0,v1,weights)

@nps.broadcast_define( (('N',3,), ('N',3,), ('N',)),
                       (3,3), )
def _align_procrustes_vectors_R01(v0, v1, weights):

    v0 = nps.transpose(v0)
    v1 = nps.transpose(v1)

    # I process Mt instead of M to not need to transpose anything later, and to
    # end up with contiguous-memory results
    Mt = nps.matmult( v0*weights, nps.transpose(v1) )
    V,S,Ut = np.linalg.svd(Mt)

    # I look at the second-lowest singular value. One 0 singular value is OK;
    # two isn't
    if S[-2] < 1e-12:
        # Poorly-defined problem. Return error
        return np.zeros((3,3), dtype=float)

    R = nps.matmult(V, Ut)

    # det(R) is now +1 or -1. If it's -1, then this contains a mirror, and thus
    # is not a physical rotation. I compensate by negating the least-important
    # pair of singular vectors
    if np.linalg.det(R) < 0:
        V[:,2] *= -1
        R = nps.matmult(V, Ut)

    return R


def align_procrustes_points_Rt01(p0, p1, weights=None):
    r"""Compute a rigid transformation to align two point clouds

SYNOPSIS

    print(points0.shape)
    ===>
    (100,3)

    print(points1.shape)
    ===>
    (100,3)

    Rt01 = mrcal.align_procrustes_points_Rt01(points0, points1)

    print( np.sum(nps.norm2(mrcal.transform_point_Rt(Rt01, points1) -
                            points0)) )
    ===>
    [The fit error from applying the optimal transformation. If the two point
     clouds match up, this will be small]

Given two sets of 3D points in numpy arrays of shape (N,3), we find the optimal
rotation, translation to align these sets of points. This is done with a
well-known direct method. See:

- https://en.wikipedia.org/wiki/Orthogonal_Procrustes_problem
- https://en.wikipedia.org/wiki/Kabsch_algorithm

We return a transformation that minimizes the sum 2-norm of the misalignment:

    cost = sum( norm2( w[i] (a[i] - transform(b[i])) ))

We return an Rt transformation to map points in set 1 to points in set 0.

At least 3 sets of points must be passed-in to produce a non-singular result. If
a set of inputs produces a singular result, a (4,3) array of 0 is returned. This
is not a valid Rt transform, and is used to signal an error.

A similar computation can be performed to instead align a set of UNIT VECTORS to
compute an optimal rotation matrix R by calling align_procrustes_vectors_R01().

ARGUMENTS

- p0: an array of shape (..., N, 3). Each row is a point in the coordinate
  system we're transforming TO

- p1: an array of shape (..., N, 3). Each row is a point in the coordinate
  system we're transforming FROM

- weights: optional array of shape (..., N). Specifies the relative weight of
  each point. If omitted, all the given points are weighted equally

RETURNED VALUES

The Rt transformation in an array of shape (4,3). We return the optimal
transformation to align the given point clouds. The transformation maps points
TO coord system 0 FROM coord system 1.

If the inputs were singular or insufficient, a (4,3) array of 0 is returned.
This is not a valid Rt transform, and is used to signal an error.

    """
    # Use the python implementation when the data isn't contiguous
    if not (p0.flags['C_CONTIGUOUS'] and \
            p1.flags['C_CONTIGUOUS']):
        return _align_procrustes_points_Rt01_python(p0, p1, weights=weights)

    if weights is None:
        return mrcal._poseutils_npsp._align_procrustes_points_Rt01_noweights(p0, p1)
    else:
        return mrcal._poseutils_npsp._align_procrustes_points_Rt01_weights(p0, p1, weights)

def align_procrustes_vectors_R01(v0, v1, weights=None):
    r"""Compute a rotation to align two sets of direction vectors

SYNOPSIS

    print(vectors0.shape)
    ===>
    (100,3)

    print(vectors1.shape)
    ===>
    (100,3)

    R01 = mrcal.align_procrustes_vectors_R01(vectors0, vectors1)

    print( np.mean(1. - nps.inner(mrcal.rotate_point_R(R01, vectors1),
                                  vectors0)) )
    ===>
    [The fit error from applying the optimal rotation. If the two sets of
     vectors match up, this will be small]

Given two sets of normalized direction vectors in 3D (stored in numpy arrays of
shape (N,3)), we find the optimal rotation to align them. This is done with a
well-known direct method. See:

- https://en.wikipedia.org/wiki/Orthogonal_Procrustes_problem
- https://en.wikipedia.org/wiki/Kabsch_algorithm

We return a rotation that minimizes the weighted sum of the cosine of the
misalignment:

    cost = -sum( w[i] inner(a[i], rotate(b[i])) )

We return a rotation to map vectors in set 1 to vectors in set 0.

At least 2 sets of vectors must be passed-in to produce a non-singular result.
If a set of inputs produces a singular result, a (3,3) array of 0 is returned.
This is not a valid rotation, and is used to signal an error.

A similar computation can be performed to instead align a set of POINTS to
compute an optimal transformation Rt by calling align_procrustes_points_Rt01().

ARGUMENTS

- v0: an array of shape (..., N, 3). Each row is a vector in the coordinate
  system we're transforming TO. These are assumed normalized.

- v1: an array of shape (..., N, 3). Each row is a vector in the coordinate
  system we're transforming FROM. These are assumed normalized.

- weights: optional array of shape (..., N). Specifies the relative weight of
  each vector. If omitted, everything is weighted equally

RETURNED VALUES

The rotation matrix in an array of shape (3,3). We return the optimal rotation
to align the given vector sets. The rotation maps vectors TO coord system 0 FROM
coord system 1.

If the inputs were singular or insufficient, a (3,3) array of 0 is returned.
This is not a valid rotation, and is used to signal an error.

    """

    # Use the python implementation when the data isn't contiguous
    if not (v0.flags['C_CONTIGUOUS'] and \
            v1.flags['C_CONTIGUOUS']):
        return _align_procrustes_vectors_R01_python(v0, v1, weights=weights)

    if weights is None:
        return mrcal._poseutils_npsp._align_procrustes_vectors_R01_noweights(v0, v1)
    else:
        return mrcal._poseutils_npsp._align_procrustes_vectors_R01_weights(v0, v1, weights)


def sample_imager(gridn_width, gridn_height, imager_width, imager_height):
    r'''Returns regularly-sampled, gridded pixels coordinates across the imager

SYNOPSIS

    q = sample_imager( 60, 40, *model.imagersize() )

    print(q.shape)
    ===>
    (40,60,2)

Note that the arguments are given in width,height order, as is customary when
generally talking about images and indexing. However, the output is in
height,width order, as is customary when talking about matrices and numpy
arrays.

If we ask for gridding dimensions (gridn_width, gridn_height), the output has
shape (gridn_height,gridn_width,2) where each row is an (x,y) pixel coordinate.

The top-left corner is at [0,0,:]:

    sample_imager(...)[0,0] = [0,0]

The the bottom-right corner is at [-1,-1,:]:

     sample_imager(...)[            -1,           -1,:] =
     sample_imager(...)[gridn_height-1,gridn_width-1,:] =
     (imager_width-1,imager_height-1)

When making plots you probably want to call mrcal.imagergrid_using(). See the
that docstring for details.

ARGUMENTS

- gridn_width: how many points along the horizontal gridding dimension

- gridn_height: how many points along the vertical gridding dimension. If None,
  we compute an integer gridn_height to maintain a square-ish grid:
  gridn_height/gridn_width ~ imager_height/imager_width

- imager_width,imager_height: the width, height of the imager. With a
  mrcal.cameramodel object this is *model.imagersize()

RETURNED VALUES

We return an array of shape (gridn_height,gridn_width,2). Each row is an (x,y)
pixel coordinate.

    '''

    if gridn_height is None:
        gridn_height = int(round(imager_height/imager_width*gridn_width))

    w = np.linspace(0,imager_width -1,gridn_width)
    h = np.linspace(0,imager_height-1,gridn_height)
    return np.ascontiguousarray(nps.mv(nps.cat(*np.meshgrid(w,h)),
                                       0,-1))


def sample_imager_unproject(gridn_width,  gridn_height,
                            imager_width, imager_height,
                            lensmodel, intrinsics_data,
                            normalize = False):
    r'''Reports 3D observation vectors that regularly sample the imager

SYNOPSIS

    import gnuplotlib as gp
    import mrcal

    ...

    Nwidth  = 60
    Nheight = 40

    # shape (Nheight,Nwidth,3)
    v,q = \
        mrcal.sample_imager_unproject(Nw, Nh,
                                      *model.imagersize(),
                                      *model.intrinsics())

    # shape (Nheight,Nwidth)
    f = interesting_quantity(v)

    gp.plot(f,
            tuplesize = 3,
            ascii     = True,
            using     = mrcal.imagergrid_using(model.imagersize, Nw, Nh),
            square    = True,
            _with     = 'image')

This is a utility function used by functions that evalute some interesting
quantity for various locations across the imager. Grid dimensions and lens
parameters are passed in, and the grid points and corresponding unprojected
vectors are returned. The unprojected vectors are unique only up-to-length, and
the returned vectors aren't normalized by default. If we want them to be
normalized, pass normalize=True.

This function has two modes of operation:

- One camera. lensmodel is a string, and intrinsics_data is a 1-dimensions numpy
  array. With a mrcal.cameramodel object together these are *model.intrinsics().
  We return (v,q) where v is a shape (Nheight,Nwidth,3) array of observation
  vectors, and q is a (Nheight,Nwidth,2) array of corresponding pixel
  coordinates (the grid returned by sample_imager())

- Multiple cameras. lensmodel is a list or tuple of strings; intrinsics_data is
  an iterable of 1-dimensional numpy arrays (a list/tuple or a 2D array). We
  return the same q as before (only one camera is gridded), but the unprojected
  array v has shape (Ncameras,Nheight,Nwidth,3) where Ncameras is the leading
  dimension of lensmodel. The gridded imager appears in camera0: v[0,...] =
  unproject(q)

ARGUMENTS

- gridn_width: how many points along the horizontal gridding dimension

- gridn_height: how many points along the vertical gridding dimension. If None,
  we compute an integer gridn_height to maintain a square-ish grid:
  gridn_height/gridn_width ~ imager_height/imager_width

- imager_width,imager_height: the width, height of the imager. With a
  mrcal.cameramodel object this is *model.imagersize()

- lensmodel, intrinsics_data: the lens parameters. With a single camera,
  lensmodel is a string, and intrinsics_data is a 1-dimensions numpy array; with
  a mrcal.cameramodel object together these are *model.intrinsics(). With
  multiple cameras, lensmodel is a list/tuple of strings. And intrinsics_data is
  an iterable of 1-dimensional numpy arrays (a list/tuple or a 2D array).

- normalize: optional boolean defaults to False. If True: normalize the output
  vectors

RETURNED VALUES

We return a tuple:

- v: the unprojected vectors. If we have a single camera this has shape
  (Nheight,Nwidth,3). With multiple cameras this has shape
  (Ncameras,Nheight,Nwidth,3). These are NOT normalized by default. To get
  normalized vectors, pass normalize=True

- q: the imager-sampling grid. This has shape (Nheight,Nwidth,2) regardless of
  how many cameras were given (we always sample just one camera). This is what
  sample_imager() returns

    '''

    def is_list_or_tuple(l):
        return isinstance(l,list) or isinstance(l,tuple)


    # shape: (Nheight,Nwidth,2). Contains (x,y) rows
    grid = sample_imager(gridn_width, gridn_height, imager_width, imager_height)

    if is_list_or_tuple(lensmodel):
        # shape: Ncameras,Nwidth,Nheight,3
        return np.array([mrcal.unproject(np.ascontiguousarray(grid),
                                         lensmodel[i],
                                         intrinsics_data[i],
                                         normalize = normalize) \
                         for i in range(len(lensmodel))]), \
               grid
    else:
        # shape: Nheight,Nwidth,3
        return \
            mrcal.unproject(np.ascontiguousarray(grid),
                            lensmodel, intrinsics_data,
                            normalize = normalize), \
            grid


def hypothesis_board_corner_positions(icam_intrinsics = None,
                                      idx_inliers     = None,
                                      **optimization_inputs):
    r'''Reports the 3D chessboard points observed by a camera at calibration time

SYNOPSIS

    model = mrcal.cameramodel("xxx.cameramodel")

    optimization_inputs = model.optimization_inputs()

    # shape (Nobservations, Nheight, Nwidth, 3)
    pcam = mrcal.hypothesis_board_corner_positions(**optimization_inputs)[0]

    i_intrinsics = \
      optimization_inputs['indices_frame_camintrinsics_camextrinsics'][:,1]

    # shape (Nobservations,1,1,Nintrinsics)
    intrinsics = nps.mv(optimization_inputs['intrinsics'][i_intrinsics],-2,-4)

    optimization_inputs['observations_board'][...,:2] = \
        mrcal.project( pcam,
                       optimization_inputs['lensmodel'],
                       intrinsics )

    # optimization_inputs now contains perfect, noiseless board observations

    x = mrcal.optimizer_callback(**optimization_inputs)[1]
    print(nps.norm2(x[:mrcal.num_measurements_boards(**optimization_inputs)]))
    ==>
    0

The optimization routine generates hypothetical observations from a set of
parameters being evaluated, trying to match these hypothetical observations to
real observations. To facilitate analysis, this routine returns these
hypothetical coordinates of the chessboard corners being observed. This routine
reports the 3D points in the coordinate system of the observing camera.

The hypothetical points are constructed from

- The calibration object geometry
- The calibration object-reference transformation in
  optimization_inputs['rt_ref_frame']
- The camera extrinsics (reference-camera transformation) in
  optimization_inputs['rt_cam_ref']
- The table selecting the camera and calibration object frame for each
  observation in
  optimization_inputs['indices_frame_camintrinsics_camextrinsics']

ARGUMENTS

- icam_intrinsics: optional integer specifying which intrinsic camera in the
  optimization_inputs we're looking at. If omitted (or None), we report
  camera-coordinate points for all the cameras

- idx_inliers: optional numpy array of booleans of shape
  (Nobservations,object_height,object_width) to select the outliers manually. If
  omitted (or None), the outliers are selected automatically: idx_inliers =
  observations_board[...,2] > 0. This argument is available to pick common
  inliers from two separate solves.

- **optimization_inputs: a dict() of arguments passable to mrcal.optimize() and
  mrcal.optimizer_callback(). We use the geometric data. This dict is obtainable
  from a cameramodel object by calling cameramodel.optimization_inputs()

RETURNED VALUE

- An array of shape (Nobservations, Nheight, Nwidth, 3) containing the
  coordinates (in the coordinate system of each camera) of the chessboard
  corners, for ALL the cameras. These correspond to the observations in
  optimization_inputs['observations_board'], which also have shape
  (Nobservations, Nheight, Nwidth, 3)

- An array of shape (Nobservations_thiscamera, Nheight, Nwidth, 3) containing
  the coordinates (in the camera coordinate system) of the chessboard corners,
  for the particular camera requested in icam_intrinsics. If icam_intrinsics is
  None: this is the same array as the previous returned value

- an (N,3) array containing camera-frame 3D points observed at calibration time,
  and accepted by the solver as inliers. This is a subset of the 2nd returned
  array.

- an (N,3) array containing camera-frame 3D points observed at calibration time,
  but rejected by the solver as outliers. This is a subset of the 2nd returned
  array.

    '''

    observations_board = optimization_inputs.get('observations_board')
    if observations_board is None:
        raise Exception("No board observations available")

    indices_frame_camintrinsics_camextrinsics = \
        optimization_inputs['indices_frame_camintrinsics_camextrinsics']

    object_width_n      = observations_board.shape[-2]
    object_height_n     = observations_board.shape[-3]
    object_spacing      = optimization_inputs['calibration_object_spacing']
    calobject_warp      = optimization_inputs.get('calobject_warp')
    # shape (Nh,Nw,3)
    full_object         = mrcal.ref_calibration_object(object_width_n,
                                                       object_height_n,
                                                       object_spacing,
                                                       calobject_warp = calobject_warp)
    Rt_ref_frame = \
        mrcal.Rt_from_rt( optimization_inputs['rt_ref_frame'] )\
        [ indices_frame_camintrinsics_camextrinsics[:,0] ]
    Rt_cam_ref = \
        nps.glue( mrcal.identity_Rt(),
                  mrcal.Rt_from_rt(optimization_inputs['rt_cam_ref']),
                  axis = -3 ) \
        [ indices_frame_camintrinsics_camextrinsics[:,2]+1 ]

    Rt_cam_frame = mrcal.compose_Rt( Rt_cam_ref,
                                     Rt_ref_frame )

    p_cam_calobjects = \
        mrcal.transform_point_Rt(nps.mv(Rt_cam_frame,-3,-5), full_object)

    # shape (Nobservations,Nheight,Nwidth)
    if idx_inliers is None:
        idx_inliers = observations_board[...,2] > 0
    idx_outliers = ~idx_inliers

    if icam_intrinsics is None:
        return                                       \
            p_cam_calobjects,                        \
            p_cam_calobjects,                        \
            p_cam_calobjects[idx_inliers,      ...], \
            p_cam_calobjects[idx_outliers,     ...]


    # The user asked for a specific camera. Separate out its data

    # shape (Nobservations,)
    idx_observations = indices_frame_camintrinsics_camextrinsics[:,1]==icam_intrinsics

    idx_inliers [~idx_observations] = False
    idx_outliers[~idx_observations] = False

    return                                       \
        p_cam_calobjects,                        \
        p_cam_calobjects[idx_observations, ...], \
        p_cam_calobjects[idx_inliers,      ...], \
        p_cam_calobjects[idx_outliers,     ...]


def _splined_stereographic_domain(lensmodel):
    r'''Return the stereographic domain for splined-stereographic lens models

SYNOPSIS

    model = mrcal.cameramodel(model_filename)

    lensmodel = model.intrinsics()[0]

    domain_contour = mrcal._splined_stereographic_domain(lensmodel)

Splined stereographic models are defined by a splined surface. This surface is
indexed by normalized stereographic-projected points. This surface is defined in
some finite area, and this function reports a piecewise linear contour reporting
this region.

This function only makes sense for splined stereographic models.

RETURNED VALUE

An array of shape (N,2) containing a contour representing the projection domain.

    '''

    if not re.match('LENSMODEL_SPLINED_STEREOGRAPHIC', lensmodel):
        raise Exception(f"This only makes sense with splined models. Input uses {lensmodel}")

    ux,uy = mrcal.knots_for_splined_models(lensmodel)
    # shape (Ny,Nx,2)
    u = np.ascontiguousarray(nps.mv(nps.cat(*np.meshgrid(ux,uy)), 0, -1))

    meta = mrcal.lensmodel_metadata_and_config(lensmodel)
    if meta['order'] == 2:
        # spline order is 3. The valid region is 1/2 segments inwards from the
        # outer contour
        return \
            nps.glue( (u[0,1:-2] + u[1,1:-2]) / 2.,
                      (u[0,-2] + u[1,-2] + u[0,-1] + u[1,-1]) / 4.,

                      (u[1:-2, -2] + u[1:-2, -1]) / 2.,
                      (u[-2,-2] + u[-1,-2] + u[-2,-1] + u[-1,-1]) / 4.,

                      (u[-2, -2:1:-1] + u[-1, -2:1:-1]) / 2.,
                      (u[-2, 1] + u[-1, 1] + u[-2, 0] + u[-1, 0]) / 4.,

                      (u[-2:0:-1, 0] +u[-2:0:-1, 1]) / 2.,
                      (u[0, 0] +u[0, 1] + u[1, 0] +u[1, 1]) / 4.,

                      (u[0,1] + u[1,1]) / 2.,
                      axis = -2 )

    elif meta['order'] == 3:
        # spline order is 3. The valid region is the outer contour, leaving one
        # knot out
        return \
            nps.glue( u[1,1:-2], u[1:-2, -2], u[-2, -2:1:-1], u[-2:0:-1, 1],
                      axis=-2 )
    else:
        raise Exception("I only support cubic (order==3) and quadratic (order==2) models")


def polygon_difference(positive, negative):
    r'''Return the difference of two closed polygons

SYNOPSIS

    import numpy as np
    import numpysane as nps
    import gnuplotlib as gp

    A = np.array(((-1,-1),( 1,-1),( 1, 1),(-1, 1),(-1,-1)))
    B = np.array(((-.1,-1.1),( .1,-1.1),( .1, 1.1),(-.1, 1.1),(-.1,-1.1)))

    diff = mrcal.polygon_difference(A, B)

    gp.plot( (A, dict(legend = 'A', _with = 'lines')),
             (B, dict(legend = 'B', _with = 'lines')),
             *[ ( r, dict( _with     = 'filledcurves closed fillcolor "red"',
                           legend    = 'difference'))
                for r in diff],
             tuplesize = -2,
             square    = True,
             wait      = True)

Given two polygons specified as a point sequence in arrays of shape (N,2) this
function computes the topological difference: all the regions contained in the
positive polygon, but missing in the negative polygon. The result could be
empty, or it could contain any number of disconnected polygons, so a list of
polygons is returned. Each of the constituent resulting polygons is guaranteed
to not have holes. If any holes are found when computing the difference, we cut
apart the resulting shape until no holes remain.

ARGUMENTS

- positive: a polygon specified by a sequence of points in an array of shape
  (N,2). The resulting difference describes regions contained inside the
  positive polygon

- negative: a polygon specified by a sequence of points in an array of shape
  (N,2). The resulting difference describes regions outside the negative polygon

RETURNED VALUE

A list of arrays of shape (N,2). Each array in the list describes a hole-free
polygon as a sequence of points. The difference is a union of all these
constituent polygons. This list could have 0 elements (empty difference) or N
element (difference consists of N separate polygons)

    '''

    from shapely.geometry import Polygon,MultiPolygon,GeometryCollection,LineString
    import shapely.ops


    diff = Polygon(positive).difference(Polygon(negative))
    if isinstance(diff, (MultiPolygon,GeometryCollection)):
        diff = list(diff.geoms)
    elif isinstance(diff, Polygon):
        diff = [diff]
    else:
        raise Exception(f"I only know how to deal with MultiPolygon or Polygon, but instead got type '{type(diff)}")

    def split_polygon_to_remove_holes(p):
        if not isinstance(p, Polygon):
            raise Exception(f"Expected a 'Polygon' type, but got {type(p)}")

        if not (p.interiors and len(p.interiors)):
            # No hole. Return the coords, if they exist
            try:
                coords = p.exterior.coords
                if len(coords) == 0:
                    return []
                return [np.array(coords)]
            except:
                return []

        # There's a hole! We need to split this polygon. I cut the polygon by a
        # line between the centroid and some vertex. Which one doesn't matter; I
        # keep trying until some cut works
        hole = p.interiors[0]
        for i in range(0,len(hole.coords)):

            l0 = np.array((hole.centroid))
            l1 = np.array((hole.coords[i]))
            l0,l1 = (l1 + 100*(l0-l1)),(l0 + 100*(l1-l0))
            line = LineString( (l0,l1) )

            s = shapely.ops.split(p, line)
            if len(s) > 1:
                # success. split into multiple pieces. I process each one
                # recursively, and I'm done. I return a flattened list
                return [subpiece for piece in s for subpiece in split_polygon_to_remove_holes(piece)]
            # Split didn't work. Try the next vertex

        print("WARNING: Couldn't split the region. Ignoring",
              file = sys.stderr)
        return []

    return \
        [subpiece for p in diff for subpiece in split_polygon_to_remove_holes(p)]


def _densify_polyline(p, spacing):
    r'''Returns the input polyline, but resampled more densely
    The input and output polylines are a numpy array of shape (N,2). The output
    is resampled such that each input point is hit, but each linear segment is
    also sampled with at least the given spacing

    '''

    if p is None or p.size == 0:
        return p

    p1 = np.array(p[0,:], dtype=p.dtype)

    for i in range(1,len(p)):
        a = p[i-1,:]
        b = p[i,  :]
        d = b-a
        l = nps.mag(d)

        # A hacky method of rounding up
        N = int(l/spacing - 1e-6 + 1.)

        for j in range(N):
            p1 = nps.glue(p1,
                          float(j+1) / N * d + a,
                          axis=-2)
    return p1


# mrcal.shellquote is either pipes.quote or shlex.quote, depending on
# python2/python3
try:
    import pipes
    shellquote = pipes.quote
except:
    # python3 puts this into a different module
    import shlex
    shellquote = shlex.quote

def mapping_file_framenocameraindex(*files_per_camera):
    r'''Parse image filenames to get the frame numbers

SYNOPSIS

    mapping = \
      mapping_file_framenocameraindex( ('img5-cam2.jpg', 'img6-cam2.jpg'),
                                       ('img6-cam3.jpg', 'img7-cam3.jpg'),)

    print(mapping)
    ===>
    { 'frame5-cam2.jpg': (5, 0),
      'frame6-cam2.jpg': (6, 0),
      'frame6-cam3.jpg': (6, 1),
      'frame7-cam3.jpg': (7, 1) }


Prior to this call we already applied a glob to some images, so we already know
which images belong to which camera. This function further classifies the images
to find the frame number of each image. This is done by looking at the filenames
of images in each camera, removing common prefixes and suffixes, and using the
central varying filename component as the frame number. This varying component
should be numeric. If it isn't and we have multiple cameras, then we barf. If it
isn't, but we only have one camera, we fallback on sequential frame numbers.

If we have just one image for a camera, I can't tell what is constant in the
filenames, so I return framenumber=0.

ARGUMENTS

- *files_per_camera: one argument per camera. Each argument is a list of strings
   of filenames of images observed by that camera

RETURNED VALUES

We return a dict from filenames to (framenumber, cameraindex) tuples. The
"cameraindex" is a sequential index counting up from 0. cameraindex==0
corresponds to files_per_camera[0] and so on.

The "framenumber" may not be sequential OR starting from 0: this comes directly
from the filename.

    '''

    i_empty = [i for i in range(len(files_per_camera)) if len(files_per_camera[i]) == 0]
    if len(i_empty) > 0:
        raise Exception("These camera globs matched no files: {}".format(i_empty))


    def get_longest_leading_trailing_substrings(strings):
        r'''Given a list of strings, returns the length of the longest leading and
        trailing substring common to all the strings

        Main use case is to take in strings such as

          a/b/c/frame001.png
          a/b/c/frame002.png
          a/b/c/frame003.png

        and return ("a/b/c/frame00", ".png")

        '''

        # These feel inefficient, especially being written in python. There's
        # probably some built-in primitive I'm not seeing
        def longest_leading_substring(a,b):
            for i in range(len(a)):
                if i >= len(b) or a[i] != b[i]:
                    return a[:i]
            return a
        def longest_trailing_substring(a,b):
            for i in range(len(a)):
                if i >= len(b) or a[-i-1] != b[-i-1]:
                    if i == 0:
                        return ''
                    return a[-i:]
            return a

        if not strings:
            return (None,None)

        leading  = strings[0]
        trailing = strings[0]

        for s in strings[1:]:
            leading  = longest_leading_substring (leading,s)
            trailing = longest_trailing_substring(trailing,s)
        return leading,trailing

    def pull_framenumbers(files):

        if len(files) == 1:
            # special case where only one file is given. In this case I can't
            # tell where the frame number is, but I don't really care. I just
            # say that the frame number is 0
            return [0]

        leading,trailing = get_longest_leading_trailing_substrings(files)
        Nleading  = len(leading)
        Ntrailing = len(trailing)

        # I now have leading and trailing substrings. I make sure that all the stuff
        # between the leading and trailing strings is numeric

        # needed because I want s[i:-0] to mean s[i:], but that doesn't work, but
        # s[i:None] does
        Itrailing = -Ntrailing if Ntrailing > 0 else None
        for f in files:
            if not re.match("^[0-9]+$", f[Nleading:Itrailing]):
                raise Exception(("Image filenames MUST be of the form 'something..number..something'\n" +   \
                                 "where the somethings are common to all the filenames. File '{}'\n" + \
                                 "has a non-numeric middle: '{}'. The somethings are: '{}' and '{}'\n" + \
                                 "Did you forget to pass globs for each camera separately?"). \
                                format(f, f[Nleading:Itrailing],
                                       leading, trailing))

        # Alrighty. The centers are all numeric. I gather all the digits around the
        # centers, and I'm done
        m = re.match("^(.*?)([0-9]*)$", leading)
        if m:
            pre_numeric = m.group(2)
        else:
            pre_numeric = ''

        m = re.match("^([0-9]*)(.*?)$", trailing)
        if m:
            post_numeric = m.group(1)
        else:
            post_numeric = ''

        return [int(pre_numeric + f[Nleading:Itrailing] + post_numeric) for f in files]




    Ncameras = len(files_per_camera)
    mapping = {}
    for icamera in range(Ncameras):
        try:
            framenumbers = pull_framenumbers(files_per_camera[icamera])
        except:
            # If we couldn't parse out the frame numbers, but there's only one
            # camera, then I just use a sequential list of integers. Since it
            # doesn't matter
            if Ncameras == 1:
                framenumbers = range(len(files_per_camera[icamera]))
            else:
                raise
        if framenumbers is not None:
            mapping.update(zip(files_per_camera[icamera], [(iframe,icamera) for iframe in framenumbers]))
    return mapping


def close_contour(c):
    r'''Close a polyline, if it isn't already closed

SYNOPSIS

    print( a.shape )
    ===>
    (5, 2)

    print( a[0] )
    ===>
    [844 204]

    print( a[-1] )
    ===>
    [886 198]

    b = mrcal.close_contour(a)

    print( b.shape )
    ===>
    (6, 2)

    print( b[0] )
    ===>
    [844 204]

    print( b[-2:] )
    ===>
    [[886 198]
     [844 204]]

This function works with polylines represented as arrays of shape (N,2). The
polygon represented by such a polyline is "closed" if its first and last points
sit at the same location. This function ingests a polyline, and returns the
corresponding, closed polygon. If the first and last points of the input match,
the input is returned. Otherwise, the first point is appended to the end, and
this extended polyline is returned.

None is accepted as an empty polygon: we return None.

ARGUMENTS

- c: an array of shape (N,2) representing the polyline to be closed. None and
  arrays of size 0 are accepted as special cases ("unknown" and "empty"
  regions, respectively). All arrays of size 0 are transformed into an array of
  shape (0,2), as expected by the rest of the code

RETURNED VALUE

An array of shape (N,2) representing the closed polygon. The input is returned
if the input was None or has shape (0,2)

    '''
    if c is None:   return None
    if c.size == 0: return np.zeros((0,2), dtype=float)

    if np.linalg.norm( c[0,:] - c[-1,:]) < 1e-6:
        return c
    return nps.glue(c, c[0,:], axis=-2)



def _plot_options_state_meas_boundary(f,optimization_inputs,arg0):
    if arg0 is not None: i0 = f(arg0,**optimization_inputs)
    else:                i0 = f(     **optimization_inputs)
    if i0 is None:
        return

    what = f.__name__.replace('state_index_','').replace('measurement_index_','')
    yield from \
        (f'arrow from {i0}, graph 0 to {i0}, graph 1 nohead',
         f'label "{what}" at {i0},graph 0 left front offset 0,character 2 boxed')

def plotoptions_state_boundaries(**optimization_inputs):
    r'''Return the 'set' plot options for gnuplotlib to show the state boundaries

SYNOPSIS

    import numpy as np
    import gnuplotlib as gp
    import mrcal

    model               = mrcal.cameramodel('xxx.cameramodel')
    optimization_inputs = model.optimization_inputs()

    J = mrcal.optimizer_callback(**optimization_inputs)[2]

    gp.plot( np.sum(np.abs(J.toarray()), axis=-2),
             _set = mrcal.plotoptions_state_boundaries(**optimization_inputs) )

    # a plot pops up showing the magnitude of the effects of each element of the
    # packed state (as seen by the optimizer), with boundaries between the
    # different state variables denoted

When plotting the state vector (or anything relating to it, such as rows of the
Jacobian), it is usually very useful to infer at a glance the meaning of each
part of the plot. This function returns a list of 'set' directives passable to
gnuplotlib that show the boundaries inside the state vector.

ARGUMENTS

**optimization_inputs: a dict() of arguments passable to mrcal.optimize() and
mrcal.optimizer_callback(). These define the full optimization problem, and can
be obtained from the optimization_inputs() method of mrcal.cameramodel

RETURNED VALUE

A list of 'set' directives passable as plot options to gnuplotlib

    '''
    return \
        list(_plot_options_state_meas_boundary(mrcal.state_index_intrinsics,     optimization_inputs, 0)) + \
        list(_plot_options_state_meas_boundary(mrcal.state_index_extrinsics,     optimization_inputs, 0)) + \
        list(_plot_options_state_meas_boundary(mrcal.state_index_frames,         optimization_inputs, 0)) + \
        list(_plot_options_state_meas_boundary(mrcal.state_index_points,         optimization_inputs, 0)) + \
        list(_plot_options_state_meas_boundary(mrcal.state_index_calobject_warp, optimization_inputs, None))


def plotoptions_measurement_boundaries(**optimization_inputs):
    r'''Return the 'set' plot options for gnuplotlib to show the measurement boundaries

SYNOPSIS

    import numpy as np
    import gnuplotlib as gp
    import mrcal

    model               = mrcal.cameramodel('xxx.cameramodel')
    optimization_inputs = model.optimization_inputs()

    x = mrcal.optimizer_callback(**optimization_inputs)[1]

    gp.plot( np.abs(x),
             _set = mrcal.plotoptions_measurement_boundaries(**optimization_inputs) )

    # a plot pops up showing the magnitude of each measurement, with boundaries
    # between the different measurements denoted

When plotting the measurement vector (or anything relating to it, such as
columns of the Jacobian), it is usually very useful to infer at a glance the
meaning of each part of the plot. This function returns a list of 'set'
directives passable to gnuplotlib that show the boundaries inside the
measurement vector.

ARGUMENTS

**optimization_inputs: a dict() of arguments passable to mrcal.optimize() and
mrcal.optimizer_callback(). These define the full optimization problem, and can
be obtained from the optimization_inputs() method of mrcal.cameramodel

RETURNED VALUE

A list of 'set' directives passable as plot options to gnuplotlib

    '''
    return \
        list(_plot_options_state_meas_boundary(mrcal.measurement_index_boards,         optimization_inputs, 0)) + \
        list(_plot_options_state_meas_boundary(mrcal.measurement_index_points,         optimization_inputs, 0)) + \
        list(_plot_options_state_meas_boundary(mrcal.measurement_index_regularization, optimization_inputs, None))


def ingest_packed_state(b_packed,
                        **optimization_inputs):
    r'''Read a given packed state into optimization_inputs

SYNOPSIS

    # A simple gradient check

    model               = mrcal.cameramodel('xxx.cameramodel')
    optimization_inputs = model.optimization_inputs()

    b0,x0,J = mrcal.optimizer_callback(no_factorization = True,
                                       **optimization_inputs)[:3]

    db = np.random.randn(len(b0)) * 1e-9

    mrcal.ingest_packed_state(b0 + db,
                              **optimization_inputs)

    x1 = mrcal.optimizer_callback(no_factorization = True,
                                  no_jacobian      = True,
                                  **optimization_inputs)[1]

    dx_observed  = x1 - x0
    dx_predicted = nps.inner(J, db_packed)

This is the converse of mrcal.optimizer_callback(). One thing
mrcal.optimizer_callback() does is to convert the expanded (intrinsics,
extrinsics, ...) arrays into a 1-dimensional scaled optimization vector
b_packed. mrcal.ingest_packed_state() allows updates to b_packed to be absorbed
back into the (intrinsics, extrinsics, ...) arrays for further evaluation with
mrcal.optimizer_callback() and others.

ARGUMENTS

- b_packed: a numpy array of shape (Nstate,) containing the input packed state

- **optimization_inputs: a dict() of arguments passable to mrcal.optimize() and
  mrcal.optimizer_callback(). The arrays in this dict are updated


RETURNED VALUE

None

    '''

    intrinsics     = optimization_inputs.get("intrinsics")
    extrinsics     = optimization_inputs.get("rt_cam_ref")
    frames         = optimization_inputs.get("rt_ref_frame")
    points         = optimization_inputs.get("points")
    calobject_warp = optimization_inputs.get("calobject_warp")

    Npoints_fixed  = optimization_inputs.get('Npoints_fixed', 0)

    Nvars_intrinsics     = mrcal.num_states_intrinsics    (**optimization_inputs)
    Nvars_extrinsics     = mrcal.num_states_extrinsics    (**optimization_inputs)
    Nvars_frames         = mrcal.num_states_frames        (**optimization_inputs)
    Nvars_points         = mrcal.num_states_points        (**optimization_inputs)
    Nvars_calobject_warp = mrcal.num_states_calobject_warp(**optimization_inputs)

    Nvars_expected = \
        Nvars_intrinsics + \
        Nvars_extrinsics + \
        Nvars_frames     + \
        Nvars_points     + \
        Nvars_calobject_warp

    # Defaults MUST match those in OPTIMIZER_ARGUMENTS_OPTIONAL in
    # mrcal-pywrap.c. Or better yet, this whole function should
    # come from the C code instead of being reimplemented here in Python
    do_optimize_intrinsics_core        = optimization_inputs.get('do_optimize_intrinsics_core',        True)
    do_optimize_intrinsics_distortions = optimization_inputs.get('do_optimize_intrinsics_distortions', True)
    do_optimize_extrinsics             = optimization_inputs.get('do_optimize_extrinsics',             True)
    do_optimize_frames                 = optimization_inputs.get('do_optimize_frames',                 True)
    do_optimize_calobject_warp         = optimization_inputs.get('do_optimize_calobject_warp',         True)


    if b_packed.ravel().size != Nvars_expected:
        raise Exception(f"Mismatched array size: b_packed.size={b_packed.ravel().size} while the optimization problem expects {Nvars_expected}")

    b = b_packed.copy()
    mrcal.unpack_state(b, **optimization_inputs)

    if do_optimize_intrinsics_core or \
       do_optimize_intrinsics_distortions:

        ivar0 = mrcal.state_index_intrinsics(0, **optimization_inputs)
        if ivar0 is not None:
            iunpacked0,iunpacked1 = None,None # everything by default

            lensmodel    = optimization_inputs['lensmodel']
            has_core     = mrcal.lensmodel_metadata_and_config(lensmodel)['has_core']
            Ncore        = 4 if has_core else 0
            Ndistortions = mrcal.lensmodel_num_params(lensmodel) - Ncore

            if not do_optimize_intrinsics_core:
                iunpacked0 = Ncore
            if not do_optimize_intrinsics_distortions:
                iunpacked1 = -Ndistortions

            intrinsics[:, iunpacked0:iunpacked1].ravel()[:] = \
                b[ ivar0:Nvars_intrinsics ]

    if do_optimize_extrinsics:
        ivar0 = mrcal.state_index_extrinsics(0, **optimization_inputs)
        if ivar0 is not None:
            extrinsics.ravel()[:] = b[ivar0:ivar0+Nvars_extrinsics]

    if do_optimize_frames:
        ivar0 = mrcal.state_index_frames(0, **optimization_inputs)
        if ivar0 is not None:
            frames.ravel()[:] = b[ivar0:ivar0+Nvars_frames]

    if do_optimize_frames:
        ivar0 = mrcal.state_index_points(0, **optimization_inputs)
        if ivar0 is not None:
            points.ravel()[:-Npoints_fixed*3] = b[ivar0:ivar0+Nvars_points]

    if do_optimize_calobject_warp:
        ivar0 = mrcal.state_index_calobject_warp(**optimization_inputs)
        if ivar0 is not None:
            calobject_warp.ravel()[:] = b[ivar0:ivar0+Nvars_calobject_warp]


def sorted_eig(M):
    r'''Compute eigenvalues, eigenvectors; sorted results returned

SYNOPSIS

    # p is an array of shape (N,3): a point cloud. I find a normal to a fitted
    # plane
    p -= np.mean(p, axis=-2)

    l,v = mrcal.sorted_eig(nps.matmult(nps.transpose(p),p))

    n = v[:,0]

    # n is the eigenvector corresponding to the smallest eigenvalue

This function calls numpy.linalg.eig(), and sorts the output by eigenvalue, in
ascending order

ARGUMENTS

- M: the matrix to eigendecompose

RETURNED VALUE

Same as that of numpy.linalg.eig(). A tuple:

- Eigenvalues, sorted in ascending order

- Eigenvectors, sorted by the corresponding eigenvalue. Each eigenvector is
  stored in a COLUMN of this array

    '''
    l,v = np.linalg.eig(M)
    i = np.argsort(l)

    # for single matrices, I can simply do
    #   return l[i], v[:,i]
    # But to support broadcasting I need these more complex expressions
    return \
        ( np.take_along_axis(l, i, -1), \
          np.take_along_axis(v, nps.dummy(i,-2), -1) )


def _plot_arg_covariance_ellipse(q_mean, Var, what):

    # if the variance is 0, the ellipse is infinitely small, and I don't even
    # try to plot it. Gnuplot has the arguably-buggy behavior where drawing an
    # ellipse with major_diam = minor_diam = 0 plots a nominally-sized ellipse.
    if np.max(np.abs(Var)) == 0:
        return None

    l,v   = mrcal.sorted_eig(Var)
    l0,l1 = l
    v0,v1 = nps.transpose(v)

    major = np.sqrt(l0)
    minor = np.sqrt(l1)

    return \
      (q_mean[0], q_mean[1], 2*major, 2*minor, 180./np.pi*np.arctan2(v0[1],v0[0]),
       dict(_with='ellipses', tuplesize=5, legend=what))


def _plot_args_points_and_covariance_ellipse(q, what):
    q_mean  = np.mean(q,axis=-2)
    q_mean0 = q - q_mean
    Var     = np.mean( nps.outer(q_mean0,q_mean0), axis=0 )
    # Some functions assume that we're plotting _with = "dots". Look for the
    # callers if changing this
    return ( _plot_arg_covariance_ellipse(q_mean,Var, what),
             ( q, dict(_with = 'dots',
                         tuplesize = -2)) )

def measurements_board(optimization_inputs,
                       *,
                       icam_intrinsics     = None,
                       x                   = None,
                       return_observations = False,

                       # for backwards-compatibility only; same as
                       # icam_intrinsics
                       i_cam               = None):
    r'''Compute and return the chessboard measurements

SYNOPSIS

    model = mrcal.cameramodel(model_filename)

    gp.plot( mrcal.measurements_board(
               optimization_inputs = model.optimization_inputs(),
               icam_intrinsics     = icam_intrinsics ).ravel(),
             histogram = True,
             binwidth  = 0.02 )

    ... A plot pops up showing the empirical distribution of chessboard fit
    ... errors in this solve. For the given camera only

Given a calibration solve, returns the measurements of chessboard observations,
throwing out outliers and, optionally, selecting the measurements from a
specific camera. Measurements are the WEIGHTED reprojection errors in the
measurement vector.

If no chessboards are present in the solve I return arrays of appropriate shape
with N = 0

ARGUMENTS

- optimization_inputs: the optimization inputs dict passed into and returned
  from mrcal.optimize(). This describes the solved optimization problem

- icam_intrinsics: optional integer to select the camera whose measurements
  we're interested in. If omitted or None, we report the measurements for ALL
  the cameras together.

- x: optional numpy array of shape (Nmeasurements,) containing the optimization
  measurements. If omitted or None, this will be recomputed. To use a cached
  value, pass the result of mrcal.optimize(**optimization_inputs)['x'] or
  mrcal.optimizer_callback(**optimization_inputs)[1]

- return_observations: optional boolean, defaulting to False. if
  return_observations: we return a tuple (x,observations) instead of just x

RETURNED VALUES

if not return_observations:

  we return a numpy array of shape (N,2) of all the measurements. N is the number
  of pixel observations remaining after outliers and other cameras are thrown
  out

else:

  we return a tuple:

  - The same measurements array as before

  - The corresponding observation points in a numpy array of shape (N,2). These
    are a slice of observations_board[] corresponding to each measurement

    '''

    # shape (Nobservations, object_height_n, object_width_n, 3)
    observations_board = optimization_inputs.get('observations_board')
    if observations_board is None or observations_board.size == 0:
        # no board observations
        if not return_observations:
            # shape (N,2)
            return \
                np.zeros((0,2), dtype=float)
        else:
            # shape (N,2), (N,2)
            return \
                np.zeros((0,2), dtype=float), \
                np.zeros((0,2), dtype=float)



    # for backwards compatibility
    if i_cam is not None:
        icam_intrinsics = i_cam

    if x is None:
        # Flattened measurements. This is ALL the measurements: chessboard,
        # point, regularization...
        x = \
            mrcal.optimizer_callback(**optimization_inputs,
                                     no_jacobian      = True,
                                     no_factorization = True)[1]

    measurements_shape = observations_board.shape[:-1] + (2,)

    # shape (Nobservations, object_height_n, object_width_n, 2)
    imeas0  = mrcal.measurement_index_boards(0, **optimization_inputs)
    Nmeas   = mrcal.num_measurements_boards(**optimization_inputs)
    x_board = x[imeas0:imeas0+Nmeas].reshape(*measurements_shape)

    # shape (Nobservations, object_height_n, object_width_n, 3)
    indices_frame_camera = optimization_inputs['indices_frame_camintrinsics_camextrinsics'][...,:2]

    # shape (Nobservations, object_height_n, object_width_n)
    idx = np.ones( observations_board.shape[:-1], dtype=bool)

    if icam_intrinsics is not None:
        # select measurements from THIS camera
        idx[indices_frame_camera[:,1] != icam_intrinsics, ...] = False

    # select non-outliers
    idx[ observations_board[...,2] <= 0.0 ] = False

    if not return_observations:
        # shape (N,2)
        return \
            x_board[idx, ...    ]
    else:
        # shape (N,2), (N,2)
        return \
            x_board[idx, ...    ], \
            observations_board[idx, ..., :2]

def residuals_board(*args,
                    residuals = None,
                    **kwargs):
    r'''Backwards-compatibility flavor of measurements_board()

Use measurements_board() and the "x" argument. The new version uses clearer
nomenclature. We are talking about MEASUREMENTS: the optimization vector x that
contains the WEIGHTED reprojection errors

    '''
    return measurements_board(*args,
                              x = residuals,
                              **kwargs)

# For backwards compatibility
residuals_chessboard = residuals_board

def measurements_point(optimization_inputs,
                       *,
                       icam_intrinsics     = None,
                       x                   = None,
                       return_observations = False):
    r'''Compute and return the observed-point measurements

SYNOPSIS

    model = mrcal.cameramodel(model_filename)

    gp.plot( mrcal.measurements_point(
               optimization_inputs = model.optimization_inputs(),
               icam_intrinsics     = icam_intrinsics ).ravel(),
             histogram = True,
             binwidth  = 0.02 )

    ... A plot pops up showing the empirical distribution of observed-point fit
    ... errors in this solve. For the given camera only

Given a calibration solve, returns the measurements of point observations,
throwing out outliers and, optionally, selecting the measurements from a
specific camera. Measurements are the WEIGHTED reprojection errors in the
measurement vector.

If no points are present in the solve I return arrays of appropriate shape with
N = 0

ARGUMENTS

- optimization_inputs: the optimization inputs dict passed into and returned
  from mrcal.optimize(). This describes the solved optimization problem

- icam_intrinsics: optional integer to select the camera whose measurements
  we're interested in. If omitted or None, we report the measurements for ALL
  the cameras together.

- x: optional numpy array of shape (Nmeasurements,) containing the optimization
  measurements. If omitted or None, this will be recomputed. To use a cached
  value, pass the result of mrcal.optimize(**optimization_inputs)['x'] or
  mrcal.optimizer_callback(**optimization_inputs)[1]

- return_observations: optional boolean, defaulting to False. if
  return_observations: we return a tuple (x,observations) instead of just x

RETURNED VALUES

if not return_observations:

  we return a numpy array of shape (N,2) of all the measurements. N is the number
  of pixel observations remaining after outliers and other cameras are thrown
  out

else:

  we return a tuple:

  - The same measurements array as before

  - The corresponding observation points in a numpy array of shape (N,2). These
    are a slice of observations_point[] corresponding to each measurement

    '''

    # shape (Nobservations, 3)
    observations_point = optimization_inputs.get('observations_point')
    if observations_point is None or observations_point.size == 0:
        # no point observations
        if not return_observations:
            # shape (N,2)
            return \
                np.zeros((0,2), dtype=float)
        else:
            # shape (N,2), (N,2)
            return \
                np.zeros((0,2), dtype=float), \
                np.zeros((0,2), dtype=float)

    if x is None:
        # Flattened measurements. This is ALL the measurements: chessboard,
        # point, regularization...
        x = \
            mrcal.optimizer_callback(**optimization_inputs,
                                     no_jacobian      = True,
                                     no_factorization = True)[1]

    Nobservations = len(observations_point)

    imeas0 = mrcal.measurement_index_points(0, **optimization_inputs)
    Nmeas  = mrcal.num_measurements_points(**optimization_inputs)

    # shape (Nobservations, 2); each observation row is (err_x, err_y)
    x_point = x[imeas0:imeas0+Nmeas].reshape(Nobservations,2)

    indices_point_camera = optimization_inputs['indices_point_camintrinsics_camextrinsics'][...,:2]

    # shape (Nobservations,)
    idx = np.ones( (Nobservations,), dtype=bool)

    if icam_intrinsics is not None:
        # select measurements from THIS camera
        idx[indices_point_camera[:,1] != icam_intrinsics, ...] = False

    # select non-outliers
    idx[ observations_point[...,2] <= 0.0 ] = False

    if not return_observations:
        # shape (N,2)
        return \
            x_point[idx, ...    ]
    else:
        # shape (N,2), (N,2)
        return \
            x_point[idx, ...    ], \
            observations_point[idx, ..., :2]

def residuals_point(*args,
                    residuals = None,
                    **kwargs):
    r'''Backwards-compatibility flavor of measurements_point()

Use measurements_point() and the "x" argument. The new version uses clearer
nomenclature. We are talking about MEASUREMENTS: the optimization vector x that
contains the WEIGHTED reprojection errors

    '''
    return measurements_point(*args,
                              x = residuals,
                              **kwargs)

def _R_aligned_to_vector_python(v):
    r'''Reference implementation of R_aligned_to_vector in python

The main implementation is written in C in poseutils.c:

  mrcal_R_aligned_to_vector()

The two implementations are identical, with a test to verify this
    '''
    z = v/nps.mag(v)
    if np.abs(z[0]) < .9:
        x = np.array((1,0,0.))
    else:
        x = np.array((0,1,0.))
    x -= nps.inner(x,z)*z
    x /= nps.mag(x)
    y = np.cross(z,x)
    return nps.cat(x,y,z)

def write_point_cloud_as_ply(filename,
                             points,
                             *,
                             color     = None,
                             binary    = True):

    r'''Write a point cloud to a .ply file

SYNOPSIS

    print( points.shape )
    ===>
    (1000, 3)

    write_point_cloud_as_ply( "points.ply", points )

The point cloud to visualize may contain colors for each point: either as a BGR
triplet or as a single intensity value. The colors may be given either in the
trailing columns of the "points" array or in a separate "color" array.

Both binary and ascii .ply formats are supported, with the binary format the
default.

ARGUMENTS

- filename: the filename of the .ply file being written

- points: the numpy array of points being visualized. Required. One of 3 shapes
  is allowed:

  - shape (N,3) if no color information is given here; may be given in the
    separate "color" array
  - shape (N,4) if a grayscale intensity is given in the last column
  - shape (N,6) if a BGR color is given in the last 3 columns

  If color is given here, the separate "color" array must be None

- color: the numpy array of colors corresponding to each point in "points".
  Optional. If the colors are given here, they must NOT be given in "points":
  points.shape MUST be (N,3). If not None, 2 shapes are allowed:

  - shape (N,)  if we have a grayscale intensity
  - shape (N,3) if we have BGR colors

- binary: optional boolean, defaulting to True. Selects the binary or ascii .ply
  format.


RETURNED VALUE

None
'''


    if points.ndim != 2:
        raise Exception(f"points.shape must be (N,3) or (N,4) or (N,6), but points.ndim={points.ndim}")
    N,Ntrailing = points.shape

    if not (Ntrailing == 3 or \
            Ntrailing == 4 or \
            Ntrailing == 6):
        raise Exception(f"points.shape[-1] must be one of (3,4,6) to indicate no-color or grayscale or bgr colors respectively. Instead got points.shape[-1] = {points.shape[-1]}")

    if color is not None:

        if Ntrailing != 3:
            raise Exception("Both 'points' and 'color' are specifying color information. At most one of those should provide it")

        if color.ndim > 2:
            raise Exception(f"color.shape must be (N,) or (N,3), but color.ndim={color.ndim}")
        if N != color.shape[0]:
            raise Exception(f"Inconsistent point counts: len(points)={len(points)} but len(color)={len(color)}")

        if not (color.shape == (N,) or \
                color.shape == (N,3) ):
            raise Exception(f"color.shape must be (N,) or (N,3), but color.shape={color.shape}")

    else:
        # No "color" specified
        if Ntrailing == 4:
            # But it exists as grayscale in the points. Separate it into the
            # "color" array
            color  = points[:,3]
            points = points[:,:3]
        elif Ntrailing == 6:
            # But it exists as bgr in the points. Separate it into the
            # "color" array
            color  = points[:,3:]
            points = points[:,:3]



    # I now have points.shape = (N,3) and color is None or (N,) or (N,3)



    if binary:
        ply_type = 'binary_little_endian'
    else:
        ply_type = 'ascii'

    if color is not None and color.ndim == 1:
        # grayscale
        dtype = np.dtype([ ('xyz',np.float32,3), ('i', np.uint8) ])
        header_color = '''property uchar intensity
'''
    elif color is not None and color.ndim == 2:
        # bgr
        dtype = np.dtype([ ('xyz',np.float32,3), ('rgb', np.uint8, 3) ])
        header_color = '''property uchar red
property uchar green
property uchar blue
'''
    else:
        dtype = np.dtype([ ('xyz',np.float32,3), ])
        header_color = ''

    ply_header = f'''ply
format {ply_type} 1.0
element vertex {N}
property float x
property float y
property float z
{header_color}end_header
'''.encode()


    with open(filename, 'wb') as f:
        f.write(ply_header)

        if binary:
            binary_ply = np.empty( (N,),
                                   dtype = dtype)
            binary_ply['xyz'] = points.astype(np.float32)
            if color is not None and color.ndim == 1:
                # grayscale
                binary_ply['i'  ] = color .astype(np.uint8)
            elif color is not None and color.ndim == 2:
                # bgr
                binary_ply['rgb'][:,0] = color[:,2]
                binary_ply['rgb'][:,1] = color[:,1]
                binary_ply['rgb'][:,2] = color[:,0]

            binary_ply.tofile(f)
        else:
            if color is not None and color.ndim == 1:
                # grayscale
                pbgr = nps.glue(points,
                                nps.dummy(color, -1),
                                axis = -1)
                np.savetxt(f,
                           pbgr,
                           fmt      = ('%.1f','%.1f','%.1f','%d'),
                           comments = '',
                           header   = '')
            elif color is not None and color.ndim == 2:
                # bgr
                pbgr = nps.glue(points,
                                color[:,(2,)],
                                color[:,(1,)],
                                color[:,(0,)],
                                255*np.ones((N,1)),
                                axis = -1)
                np.savetxt(f,
                           pbgr,
                           fmt      = ('%.1f','%.1f','%.1f','%d','%d','%d','%d'),
                           comments = '',
                           header   = '')
            else:
                np.savetxt(f, points,
                           fmt      = ('%.1f','%.1f','%.1f',),
                           comments = '',
                           header   = '')



#!/usr/bin/python3

# Copyright (c) 2017-2023 California Institute of Technology ("Caltech"). U.S.
# Government sponsorship acknowledged. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0

'''Routines to (un)project points using any camera model

Most of these are Python wrappers around the written-in-C Python extension
module mrcal._mrcal_npsp. Most of the time you want to use this module
instead of touching mrcal._mrcal_npsp directly.

All functions are exported into the mrcal module. So you can call these via
mrcal.projections.fff() or mrcal.fff(). The latter is preferred.

'''

import numpy as np
import numpysane as nps
import sys

import mrcal


def project(v, lensmodel, intrinsics_data,
            *,
            get_gradients = False,
            out           = None):
    r'''Projects a set of 3D camera-frame points to the imager

SYNOPSIS

    # v is a (...,3) array of 3D points we're projecting
    points = mrcal.project( v,
                            lensmodel, intrinsics_data )

    ### OR ###

    m = mrcal.cameramodel(...)
    points = mrcal.project( v, *m.intrinsics() )

    # points is a (...,2) array of pixel coordinates

Given a shape-(...,3) array of points in the camera frame (x,y aligned with the
imager coords, z 'forward') and an intrinsic model, this function computes the
projection, optionally with gradients.

Projecting out-of-bounds points (beyond the field of view) returns undefined
values. Generally things remain continuous even as we move off the imager
domain. Pinhole-like projections will work normally if projecting a point behind
the camera. Splined projections clamp to the nearest spline segment: the
projection will fly off to infinity quickly since we're extrapolating a
polynomial, but the function will remain continuous.

Broadcasting is fully supported across v and intrinsics_data

ARGUMENTS

- points: array of dims (...,3); the points we're projecting

- lensmodel: a string such as

  LENSMODEL_PINHOLE
  LENSMODEL_OPENCV4
  LENSMODEL_CAHVOR
  LENSMODEL_SPLINED_STEREOGRAPHIC_order=3_Nx=16_Ny=12_fov_x_deg=100

- intrinsics: array of dims (Nintrinsics):

    (focal_x, focal_y, center_pixel_x, center_pixel_y, distortion0, distortion1,
    ...)

  The focal lengths are given in pixels.

- get_gradients: optional boolean that defaults to False. Whether we should
  compute and report the gradients. This affects what we return

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing arrays,
  specify them with the 'out' kwarg. If not get_gradients: 'out' is the one
  numpy array we will write into. Else: 'out' is a tuple of all the output numpy
  arrays. If 'out' is given, we return the same arrays passed in. This is the
  standard behavior provided by numpysane_pywrap.

RETURNED VALUE

if not get_gradients:

  we return an (...,2) array of projected pixel coordinates

if get_gradients: we return a tuple:

  - (...,2) array of projected pixel coordinates
  - (...,2,3) array of gradients of the pixel coordinates in respect to the
    input 3D point positions
  - (...,2,Nintrinsics) array of the gradients of the pixel coordinates in
    respect to the intrinsics

    '''

    # Internal function must have a different argument order so
    # that all the broadcasting stuff is in the leading arguments
    if not get_gradients:
        return mrcal._mrcal_npsp._project(v, intrinsics_data, lensmodel=lensmodel, out=out)
    return mrcal._mrcal_npsp._project_withgrad(v, intrinsics_data, lensmodel=lensmodel, out=out)


def unproject(q, lensmodel, intrinsics_data,
              *,
              normalize     = False,
              get_gradients = False,
              out           = None):
    r'''Unprojects pixel coordinates to observation vectors

SYNOPSIS

    # q is a (...,2) array of pixel observations
    v = mrcal.unproject( q,
                         lensmodel, intrinsics_data )

    ### OR ###

    m = mrcal.cameramodel(...)
    v = mrcal.unproject( q, *m.intrinsics() )

Maps a set of 2D imager points q to a set of 3D vectors in camera coordinates
that produced these pixel observations. Each 3D vector is unique only
up-to-length, and the returned vectors aren't normalized by default. The default
length of the returned vector is arbitrary, and selected for the convenience of
the implementation. Pass normalize=True to always return unit vectors.

This is the "reverse" direction, so an iterative nonlinear optimization is
performed internally to compute this result. This is much slower than
mrcal_project. For OpenCV distortions specifically, OpenCV has
cvUndistortPoints() (and cv2.undistortPoints()), but these are inaccurate and we
do not use them: https://github.com/opencv/opencv/issues/8811

Gradients are available by passing get_gradients=True. Since unproject() is
implemented as an iterative solve around project(), the unproject() gradients
are computed by manipulating the gradients reported by project() at the
solution. The reported gradients are relative to whatever unproject() is
reporting; the unprojection is unique only up-to-length, and the magnitude isn't
fixed. So the gradients may include a component in the direction of the returned
observation vector: this follows the arbitrary scaling used by unproject(). It
is possible to pass normalize=True; we then return NORMALIZED observation
vectors and the gradients of those NORMALIZED vectors. In that case, those
gradients are guaranteed to be orthogonal to the observation vector. The vector
normalization involves a bit more computation, so it isn't the default.

NOTE: THE MAGNITUDE OF THE RETURNED VECTOR CHANGES IF get_gradients CHANGES. The
reported gradients are correct relative to the output returned with
get_gradients=True. Passing normalize=True can be used to smooth this out:

    unproject(..., normalize=True)

returns the same vectors as

    unproject(..., normalize=True, get_gradients=True)[0]

Broadcasting is fully supported across q and intrinsics_data.

Models that have no gradients available cannot use mrcal_unproject() in C, but
CAN still use this mrcal.unproject() Python routine: a slower routine is
employed that uses numerical differences instead of analytical gradients.

ARGUMENTS

- q: array of dims (...,2); the pixel coordinates we're unprojecting

- lensmodel: a string such as

  LENSMODEL_PINHOLE
  LENSMODEL_OPENCV4
  LENSMODEL_CAHVOR
  LENSMODEL_SPLINED_STEREOGRAPHIC_order=3_Nx=16_Ny=12_fov_x_deg=100

- intrinsics_data: array of dims (Nintrinsics):

    (focal_x, focal_y, center_pixel_x, center_pixel_y, distortion0, distortion1,
    ...)

  The focal lengths are given in pixels.

- normalize: optional boolean defaults to False. If True: normalize the output
  vectors

- get_gradients: optional boolean that defaults to False. Whether we should
  compute and report the gradients. This affects what we return (see below). If
  not normalize, the magnitude of the reported vectors changes if get_gradients
  is turned on/off (see above)

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing arrays,
  specify them with the 'out' kwarg. If not get_gradients: 'out' is the one
  numpy array we will write into. Else: 'out' is a tuple of all the output numpy
  arrays. If 'out' is given, we return the same arrays passed in. This is the
  standard behavior provided by numpysane_pywrap.

RETURNED VALUE

if not get_gradients:

  we return an (...,3) array of unprojected observation vectors. Not normalized
  by default; see description above

if get_gradients: we return a tuple:

  - (...,3) array of unprojected observation vectors
  - (...,3,2) array of gradients of unprojected observation vectors in respect
    to pixel coordinates
  - (...,3,Nintrinsics) array of gradients of unprojected observation vectors in
    respect to the intrinsics

    '''

    def apply_normalization_to_output_with_gradients(v,dv_dq,dv_di):
        # vn = v/mag(v)
        # dvn = dv (1/mag(v)) + v d(1/mag(v))
        #     = dv( 1/mag(v) - v vt / mag^3(v) )
        #     = dv( 1/mag(v) - vn vnt / mag(v) )
        #     = dv/mag(v) ( 1 - vn vnt )

        # v has shape (...,3)
        # dv_dq has shape (...,3,2)
        # dv_di has shape (...,3,N)

        # shape (...,1)
        magv_recip = 1. / nps.dummy(nps.mag(v), -1)
        v *= magv_recip

        # shape (...,1,1)
        magv_recip = nps.dummy(magv_recip,-1)
        dv_dq *= magv_recip

        dv_dq -= nps.xchg( nps.matmult( nps.dummy(nps.xchg(dv_dq, -1,-2), -2),
                                        nps.dummy(nps.outer(v,v),-3) )[...,0,:],
                           -1, -2)

        dv_di *= magv_recip

        dv_di -= nps.xchg( nps.matmult( nps.dummy(nps.xchg(dv_di, -1,-2), -2),
                                        nps.dummy(nps.outer(v,v),-3) )[...,0,:],
                           -1, -2)


    # First, handle some trivial cases. I don't want to run the
    # optimization-based unproject() if I don't have to
    if lensmodel == 'LENSMODEL_PINHOLE' or \
       lensmodel == 'LENSMODEL_LONLAT'  or \
       lensmodel == 'LENSMODEL_LATLON'  or \
       lensmodel == 'LENSMODEL_STEREOGRAPHIC':

        if   lensmodel == 'LENSMODEL_PINHOLE':
            func = mrcal.unproject_pinhole
            always_normalized = False
        elif lensmodel == 'LENSMODEL_LONLAT':
            func = mrcal.unproject_lonlat
            always_normalized = True
        elif lensmodel == 'LENSMODEL_LATLON':
            func = mrcal.unproject_latlon
            always_normalized = True
        elif lensmodel == 'LENSMODEL_STEREOGRAPHIC':
            func = mrcal.unproject_stereographic
            always_normalized = False


        if not get_gradients:

            v = func(q, intrinsics_data, out = out)
            if normalize and not always_normalized:
                v /= nps.dummy(nps.mag(v), axis=-1)
            return v


        # shapes (...,2)
        fxy = intrinsics_data[..., :2]
        cxy = intrinsics_data[..., 2:]

        # shapes (...,3) and (...,3,2)
        v, dv_dq = \
            func(q, intrinsics_data,
                 get_gradients = True,
                 out = None if out is None else (out[0],out[1]))

        # q = f l(v) + c
        # l(v) = (q-c)/f
        #
        # dl/dv dv/df = (c-q) / f^2
        # dl/dv dv/dq = 1/f
        # -> dl/dv = 1 / ( f dv/dq )
        # -> dv/df =  (c-q) / (f^2 dl/dv) = (c-q) dv/dq / f
        #
        # dl/dv dv/dc = -1/f
        # -> dv/dc =  -1 / (f dl/dv) = -1 / (f /( f dv/dq )) = -dv/dq
        dv_di_shape = dv_dq.shape[:-1] + (4,)
        if out is None:
            dv_di = np.zeros( dv_di_shape, dtype=float)
        else:
            if not (out[2].shape[-len(dv_di_shape):] == dv_di_shape and \
                    not any(np.array(out[2].shape[:-len(dv_di_shape)]) - 1)):
                raise Exception(f"Shape of out[2] doesn't match broadcasted shape for dv_di. Wanted {dv_di_shape}, but got {out[2].shape}")
            dv_di = out[2]
            dv_di *= 0

        # dv/df
        dv_di[..., :2] += nps.dummy((cxy - q)/fxy, -2) * dv_dq
        # dv/dc
        dv_di[..., 2:] -= dv_dq

        if normalize and not always_normalized:
            apply_normalization_to_output_with_gradients(v,dv_dq,dv_di)

        return v,dv_dq,dv_di



    try:
        meta = mrcal.lensmodel_metadata_and_config(lensmodel)
    except:
        raise Exception(f"Invalid lens model '{lensmodel}': couldn't get the metadata")
    if meta['has_gradients']:

        # Main path. We have gradients.
        #
        # Internal function must have a different argument order so
        # that all the broadcasting stuff is in the leading arguments
        if not get_gradients:
            v = mrcal._mrcal_npsp._unproject(q, intrinsics_data, lensmodel=lensmodel,
                                             out=out)
            if normalize:

                # Explicitly handle nan and inf to set their normalized values
                # to 0. Otherwise I get a scary-looking warning from numpy
                i_vgood = \
                    np.isfinite(v[...,0]) * \
                    np.isfinite(v[...,1]) * \
                    np.isfinite(v[...,2])
                v[~i_vgood] = np.array((0.,0.,1.))
                v /= nps.dummy(nps.mag(v), -1)
                v[~i_vgood] = np.array((0.,0.,0.))
            return v

        # We need to report gradients
        vs = mrcal._mrcal_npsp._unproject(q, intrinsics_data, lensmodel=lensmodel)

        # I have no gradients available for unproject(), and I need to invert a
        # non-square matrix to use the gradients from project(). I deal with this
        # with a stereographic mapping
        #
        # With a simple unprojection I have    q -> v
        # Instead I now do                     q -> vs -> u -> v

        # I reproject vs, to produce a scaled v = k*vs. I'm assuming all
        # projections are central, so vs represents q just as well as v does. u
        # is a 2-vector, so dq_du is (2x2), and I can invert it
        u = mrcal.project_stereographic(vs)
        dv_du = np.zeros( vs.shape + (2,), dtype=float)
        v, dv_du = \
            mrcal.unproject_stereographic(u,
                                          get_gradients = True,
                                          out = (vs if out is None else out[0],
                                                 dv_du))

        _,dq_dv,dq_di = mrcal.project(v,
                                      lensmodel, intrinsics_data,
                                      get_gradients = True)

        # shape (..., 2,2). Square. Invertible!
        dq_du = nps.matmult( dq_dv, dv_du )

        # dv/dq = dv/du du/dq =
        #       = dv/du inv(dq/du)
        #       = transpose(inv(transpose(dq/du)) transpose(dv/du))
        dv_dq = nps.transpose(np.linalg.solve( nps.transpose(dq_du),
                                               nps.transpose(dv_du) ))
        if out is not None:
            out[1] *= 0.
            out[1] += dv_dq
            dv_dq = out[1]


        # dv/di is a bit different. I have (q,i) -> v. I want to find out
        # how moving i affects v while keeping q constant. Taylor expansion
        # of projection: q = q0 + dq/dv dv + dq/di di. q is constant so
        # dq/dv dv + dq/di di = 0 -> dv/di = - dv/dq dq/di
        dv_di = nps.matmult(dv_dq, dq_di,
                            out = None if out is None else out[2])
        dv_di *= -1.

        if normalize:
            apply_normalization_to_output_with_gradients(v,dv_dq,dv_di)

        return v, dv_dq, dv_di




    # No projection gradients implemented in C. We should get here approximately
    # never. At this time, there are no longer any projection functions that
    # have no gradients implemented. If another such model is defined, this path
    # will be used. If these see use, real gradients should be implemented
    #
    # We compute the gradients numerically. This is a reimplementation of the C
    # code. It's barely maintained, and here for legacy compatibility only
    raise("should never get here")

    if get_gradients:
        raise Exception(f"unproject(..., get_gradients=True) is unsupported for models with no gradients, such as '{lensmodel}'")

    if q is None: return q
    if q.size == 0:
        s = q.shape
        return np.zeros(s[:-1] + (3,))

    if out is not None:
        raise Exception(f"unproject(..., out) is unsupported if out is not None and we're using a model with no gradients, such as '{lensmodel}'")

    fxy = intrinsics_data[...,  :2]
    cxy = intrinsics_data[..., 2:4]

    # undistort the q, by running an optimizer

    import scipy.optimize

    # I optimize each point separately because the internal optimization
    # algorithm doesn't know that each point is independent, so if I optimized
    # it all together, it would solve a dense linear system whose size is linear
    # in Npoints. The computation time thus would be much slower than
    # linear(Npoints)
    @nps.broadcast_define( ((2,),), )
    def undistort_this(q0):

        def cost_no_gradients(vxy, *args, **kwargs):
            '''Optimization functions'''
            return \
                mrcal.project(np.array((vxy[0],vxy[1],1.)), lensmodel, intrinsics_data) - \
                q0

        # seed assuming distortions aren't there
        vxy_seed = (q0 - cxy) / fxy

        # no gradients available
        result = scipy.optimize.least_squares(cost_no_gradients, vxy_seed,
                                              '3-point')

        vxy = result.x

        # This needs to be precise; if it isn't, I barf. Shouldn't happen
        # very often
        if np.sqrt(result.cost/2.) > 1e-3:
            if not unproject.__dict__.get('already_complained'):
                sys.stderr.write("WARNING: unproject() wasn't able to precisely compute some points. Returning nan for those. Will complain just once\n")
                unproject.already_complained = True
            return np.array((np.nan,np.nan))
        return vxy

    vxy = undistort_this(q)

    # I append a 1. shape = (..., 3)
    v = nps.glue(vxy, np.ones( vxy.shape[:-1] + (1,) ), axis=-1)
    if normalize:
        v /= nps.dummy(nps.mag(v), -1)
    return v


def project_pinhole(points,
                    fxycxy = np.array((1.0, 1.0, 0.0, 0.0), dtype=float),
                    *,
                    get_gradients = False,
                    out           = None):
    r'''Projects 3D camera-frame points using a pinhole projection

SYNOPSIS

    # points is a (N,3) array of camera-coordinate-system points
    q = mrcal.project_pinhole( points, fxycxy )

    # q is now a (N,2) array of pinhole coordinates

This is a special case of mrcal.project(). Useful to represent a very simple,
very perfect lens. Wide lenses do not follow this model. Long lenses usually
more-or-less DO follow this model. See the lensmodel documentation for details:

https://mrcal.secretsauce.net/lensmodels.html#lensmodel-pinhole

Given a (N,3) array of points in the camera frame (x,y aligned with the imager
coords, z 'forward') and the parameters fxycxy, this function computes the
projection, optionally with gradients.

ARGUMENTS

- points: array of dims (...,3); the points we're projecting. This supports
  broadcasting fully, and any leading dimensions are allowed, including none

- fxycxy: optional intrinsics core. This is a shape (4,) array (fx,fy,cx,cy),
  with all elements given in units of pixels. fx and fy are the horizontal and
  vertical focal lengths, respectively. (cx,cy) are pixel coordinates
  corresponding to the projection of p = [0,0,1]. If omitted, default values are
  used: fx=fy=1.0 and cx=cy=0.0.

- get_gradients: optional boolean, defaults to False. This affects what we
  return (see below)

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing arrays,
  specify them with the 'out' kwarg. If not get_gradients: 'out' is the one
  numpy array we will write into. Else: 'out' is a tuple of all the output numpy
  arrays. If 'out' is given, we return the same arrays passed in. This is the
  standard behavior provided by numpysane_pywrap.

RETURNED VALUE

if not get_gradients: we return an (...,2) array of projected transverse
equirectangular coordinates

if get_gradients: we return a tuple:

  - (...,2) array of projected pinhole coordinates
  - (...,2,3) array of the gradients of the transverse equirectangular
    coordinates in respect to the input 3D point positions

    '''

    # Internal function must have a different argument order so
    # that all the broadcasting stuff is in the leading arguments
    if not get_gradients:
        return mrcal._mrcal_npsp._project_pinhole(points, fxycxy,
                                                  out=out)
    return mrcal._mrcal_npsp._project_pinhole_withgrad(points, fxycxy,
                                                       out=out)


def unproject_pinhole(points,
                      fxycxy = np.array((1.0, 1.0, 0.0, 0.0), dtype=float),
                      *,
                      get_gradients = False,
                      out           = None):
    r'''Unprojects 2D pixel coordinates using a pinhole projection

SYNOPSIS

    # points is a (N,2) array of imager points
    v = mrcal.unproject_pinhole( points,
                                 fxycxy )

    # v is now a (N,3) array of observation directions in the camera coordinate
    # system. v are NOT normalized

This is a special case of mrcal.unproject(). Useful to represent a very simple,
very perfect lens. Wide lenses do not follow this model. Long lenses usually
more-or-less DO follow this model. See the lensmodel documentation for details:

https://mrcal.secretsauce.net/lensmodels.html#lensmodel-pinhole

Given a (N,2) array of pinhole coordinates and the parameters fxycxy, this
function computes the inverse projection, optionally with gradients.

The vectors returned by this function are NOT normalized.

ARGUMENTS

- points: array of dims (...,2); the pinhole coordinates
  we're unprojecting. This supports broadcasting fully, and any leading
  dimensions are allowed, including none

- fxycxy: optional intrinsics core. This is a shape (4,) array (fx,fy,cx,cy),
  with all elements given in units of pixels. fx and fy are the horizontal and
  vertical focal lengths, respectively. (cx,cy) are pixel coordinates
  corresponding to the projection of p = [0,0,1]. If omitted, default values are
  used: fx=fy=1.0 and cx=cy=0.0.

- get_gradients: optional boolean, defaults to False. This affects what we
  return (see below)

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing arrays,
  specify them with the 'out' kwarg. If not get_gradients: 'out' is the one
  numpy array we will write into. Else: 'out' is a tuple of all the output numpy
  arrays. If 'out' is given, we return the same arrays passed in. This is the
  standard behavior provided by numpysane_pywrap.

RETURNED VALUE

if not get_gradients: we return an (...,3) array of unprojected observation
vectors. These are NOT normalized.

if get_gradients: we return a tuple:

  - (...,3) array of unprojected observation vectors. These are NOT normalized.
  - (...,3,2) array of the gradients of the observation vectors in respect to
    the input 2D pinhole coordinates

    '''
    if not get_gradients:
        return mrcal._mrcal_npsp._unproject_pinhole(points, fxycxy,
                                                    out=out)
    return mrcal._mrcal_npsp._unproject_pinhole_withgrad(points, fxycxy,
                                                         out=out)


def project_stereographic(points,
                          fxycxy = np.array((1.0, 1.0, 0.0, 0.0), dtype=float),
                          *,
                          get_gradients = False,
                          out           = None):
    r'''Projects a set of 3D camera-frame points using a stereographic model

SYNOPSIS

    # points is a (N,3) array of camera-coordinate-system points
    q = mrcal.project_stereographic( points )

    # q is now a (N,2) array of normalized stereographic coordinates

This is a special case of mrcal.project(). No actual lens ever follows this
model exactly, but this is useful as a baseline for other models. See the
lensmodel documentation for details:

https://mrcal.secretsauce.net/lensmodels.html#lensmodel-stereographic

Given a (N,3) array of points in the camera frame (x,y aligned with the imager
coords, z 'forward') and parameters of a perfect stereographic camera, this
function computes the projection, optionally with gradients.

The user can pass in focal length and center-pixel values. Or they can be
omitted to compute a "normalized" stereographic projection (fx = fy = 1, cx = cy
= 0).

The stereographic projection is able to represent points behind the camera, and
has only one singular observation direction: directly behind the camera, along
the optical axis.

This projection acts radially. If the observation vector v makes an angle theta
with the optical axis, then the projected point q is 2 tan(theta/2) f from the
image center.

ARGUMENTS

- points: array of dims (...,3); the points we're projecting. This supports
  broadcasting fully, and any leading dimensions are allowed, including none

- fxycxy: optional intrinsics core. This is a shape (4,) array (fx,fy,cx,cy),
  with all elements given in units of pixels. fx and fy are the horizontal and
  vertical focal lengths, respectively. (cx,cy) are pixel coordinates
  corresponding to the projection of p = [0,0,1]. If omitted, default values are
  used to specify a normalized stereographic projection : fx=fy=1.0 and
  cx=cy=0.0.

- get_gradients: optional boolean, defaults to False. This affects what we
  return (see below)

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing arrays,
  specify them with the 'out' kwarg. If not get_gradients: 'out' is the one
  numpy array we will write into. Else: 'out' is a tuple of all the output numpy
  arrays. If 'out' is given, we return the same arrays passed in. This is the
  standard behavior provided by numpysane_pywrap.

RETURNED VALUE

if not get_gradients: we return an (...,2) array of projected stereographic
coordinates

if get_gradients: we return a tuple:

  - (...,2) array of projected stereographic coordinates
  - (...,2,3) array of the gradients of the stereographic coordinates in respect
    to the input 3D point positions

    '''
    if not get_gradients:
        return mrcal._mrcal_npsp._project_stereographic(points, fxycxy,
                                                        out=out)
    return mrcal._mrcal_npsp._project_stereographic_withgrad(points, fxycxy,
                                                             out=out)


def unproject_stereographic(points,
                            fxycxy = np.array((1.0, 1.0, 0.0, 0.0), dtype=float),
                            *,
                            get_gradients = False,
                            out           = None):
    r'''Unprojects a set of 2D pixel coordinates using a stereographic model

SYNOPSIS

    # points is a (N,2) array of pixel coordinates
    v = mrcal.unproject_stereographic( points, fxycxy)

    # v is now a (N,3) array of observation directions in the camera coordinate
    # system. v are NOT normalized

This is a special case of mrcal.unproject(). No actual lens ever follows this
model exactly, but this is useful as a baseline for other models. See the
lensmodel documentation for details:

https://mrcal.secretsauce.net/lensmodels.html#lensmodel-stereographic

Given a (N,2) array of stereographic coordinates and parameters of a perfect
stereographic camera, this function computes the inverse projection, optionally
with gradients.

The user can pass in focal length and center-pixel values. Or they can be
omitted to compute a "normalized" stereographic projection (fx = fy = 1, cx = cy
= 0).

The stereographic projection is able to represent points behind the camera, and
has only one singular observation direction: directly behind the camera, along
the optical axis.

This projection acts radially. If the observation vector v makes an angle theta
with the optical axis, then the projected point q is 2 tan(theta/2) f from the
image center.

ARGUMENTS

- points: array of dims (...,2); the stereographic coordinates we're
  unprojecting. This supports broadcasting fully, and any leading dimensions are
  allowed, including none

- fxycxy: optional intrinsics core. This is a shape (4,) array (fx,fy,cx,cy),
  with all elements given in units of pixels. fx and fy are the horizontal and
  vertical focal lengths, respectively. (cx,cy) are pixel coordinates
  corresponding to the projection of p = [0,0,1]. If omitted, default values are
  used to specify a normalized stereographic projection : fx=fy=1.0 and
  cx=cy=0.0.

- get_gradients: optional boolean, defaults to False. This affects what we
  return (see below)

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing arrays,
  specify them with the 'out' kwarg. If not get_gradients: 'out' is the one
  numpy array we will write into. Else: 'out' is a tuple of all the output numpy
  arrays. If 'out' is given, we return the same arrays passed in. This is the
  standard behavior provided by numpysane_pywrap.

RETURNED VALUE

if not get_gradients: we return an (...,3) array of unprojected observation
vectors. These are NOT normalized.

if get_gradients: we return a tuple:

  - (...,3) array of unprojected observation vectors. These are NOT normalized.
  - (...,3,2) array of the gradients of the observation vectors in respect to
    the input 2D stereographic coordinates

    '''
    if not get_gradients:
        return mrcal._mrcal_npsp._unproject_stereographic(points, fxycxy,
                                                          out=out)
    return mrcal._mrcal_npsp._unproject_stereographic_withgrad(points, fxycxy,
                                                               out=out)


def project_lonlat(points,
                   fxycxy = np.array((1.0, 1.0, 0.0, 0.0), dtype=float),
                   *,
                   get_gradients = False,
                   out           = None):
    r'''Projects a set of 3D camera-frame points using an equirectangular projection

SYNOPSIS

    # points is a (N,3) array of camera-coordinate-system points
    q = mrcal.project_lonlat( points, fxycxy )

    # q is now a (N,2) array of equirectangular coordinates

This is a special case of mrcal.project(). Useful not for
representing lenses, but for describing the projection function of wide
panoramic images. Lenses do not follow this model. See the lensmodel
documentation for details:

https://mrcal.secretsauce.net/lensmodels.html#lensmodel-lonlat

Given a (N,3) array of points in the camera frame (x,y aligned with the imager
coords, z 'forward') and the parameters fxycxy, this function computes the
projection, optionally with gradients.

ARGUMENTS

- points: array of dims (...,3); the points we're projecting. This supports
  broadcasting fully, and any leading dimensions are allowed, including none

- fxycxy: optional intrinsics core. This is a shape (4,) array (fx,fy,cx,cy). fx
  and fy are the "focal lengths": they specify the angular resolution of the
  image, in pixels/radian. (cx,cy) are pixel coordinates corresponding to the
  projection of p = [0,0,1]. If omitted, default values are used to specify a
  normalized equirectangular projection : fx=fy=1.0 and cx=cy=0.0. This produces
  q = (lon,lat)

- get_gradients: optional boolean, defaults to False. This affects what we
  return (see below)

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing arrays,
  specify them with the 'out' kwarg. If not get_gradients: 'out' is the one
  numpy array we will write into. Else: 'out' is a tuple of all the output numpy
  arrays. If 'out' is given, we return the same arrays passed in. This is the
  standard behavior provided by numpysane_pywrap.

RETURNED VALUE

if not get_gradients: we return an (...,2) array of projected equirectangular
coordinates

if get_gradients: we return a tuple:

  - (...,2) array of projected equirectangular coordinates
  - (...,2,3) array of the gradients of the equirectangular coordinates in respect
    to the input 3D point positions

    '''

    # Internal function must have a different argument order so
    # that all the broadcasting stuff is in the leading arguments
    if not get_gradients:
        return mrcal._mrcal_npsp._project_lonlat(points, fxycxy,
                                                 out=out)
    return mrcal._mrcal_npsp._project_lonlat_withgrad(points, fxycxy,
                                                      out=out)


def unproject_lonlat(points,
                     fxycxy = np.array((1.0, 1.0, 0.0, 0.0), dtype=float),
                     *,
                     get_gradients = False,
                     out           = None):
    r'''Unprojects a set of 2D pixel coordinates using an equirectangular projection

SYNOPSIS

    # points is a (N,2) array of imager points
    v = mrcal.unproject_lonlat( points, fxycxy )

    # v is now a (N,3) array of observation directions in the camera coordinate
    # system. v are normalized

This is a special case of mrcal.unproject(). Useful not for
representing lenses, but for describing the projection function of wide
panoramic images. Lenses do not follow this model. See the lensmodel
documentation for details:

https://mrcal.secretsauce.net/lensmodels.html#lensmodel-lonlat

Given a (N,2) array of equirectangular coordinates and the parameters fxycxy,
this function computes the inverse projection, optionally with gradients.

The vectors returned by this function are normalized.

ARGUMENTS

- points: array of dims (...,2); the equirectangular coordinates we're
  unprojecting. This supports broadcasting fully, and any leading dimensions are
  allowed, including none

- fxycxy: optional intrinsics core. This is a shape (4,) array (fx,fy,cx,cy). fx
  and fy are the "focal lengths": they specify the angular resolution of the
  image, in pixels/radian. (cx,cy) are pixel coordinates corresponding to the
  projection of p = [0,0,1]. If omitted, default values are used to specify a
  normalized equirectangular projection : fx=fy=1.0 and cx=cy=0.0. This produces
  q = (lon,lat)

- get_gradients: optional boolean, defaults to False. This affects what we
  return (see below)

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing arrays,
  specify them with the 'out' kwarg. If not get_gradients: 'out' is the one
  numpy array we will write into. Else: 'out' is a tuple of all the output numpy
  arrays. If 'out' is given, we return the same arrays passed in. This is the
  standard behavior provided by numpysane_pywrap.

RETURNED VALUE

if not get_gradients: we return an (...,3) array of unprojected observation
vectors. These are normalized.

if get_gradients: we return a tuple:

  - (...,3) array of unprojected observation vectors. These are normalized.
  - (...,3,2) array of the gradients of the observation vectors in respect to
    the input 2D equirectangular coordinates

    '''
    if not get_gradients:
        return mrcal._mrcal_npsp._unproject_lonlat(points, fxycxy,
                                                   out=out)
    return mrcal._mrcal_npsp._unproject_lonlat_withgrad(points, fxycxy,
                                                        out=out)


def project_latlon(points,
                   fxycxy = np.array((1.0, 1.0, 0.0, 0.0), dtype=float),
                   *,
                   get_gradients = False,
                   out           = None):
    r'''Projects 3D camera-frame points using a transverse equirectangular projection

SYNOPSIS

    # points is a (N,3) array of camera-coordinate-system points
    q = mrcal.project_latlon( points, fxycxy )

    # q is now a (N,2) array of transverse equirectangular coordinates

This is a special case of mrcal.project(). Useful not for representing lenses,
but for performing stereo rectification. Lenses do not follow this model. See
the lensmodel documentation for details:

https://mrcal.secretsauce.net/lensmodels.html#lensmodel-latlon

Given a (N,3) array of points in the camera frame (x,y aligned with the imager
coords, z 'forward') and the parameters fxycxy, this function computes the
projection, optionally with gradients.

ARGUMENTS

- points: array of dims (...,3); the points we're projecting. This supports
  broadcasting fully, and any leading dimensions are allowed, including none

- fxycxy: optional intrinsics core. This is a shape (4,) array (fx,fy,cx,cy). fx
  and fy are the "focal lengths": they specify the angular resolution of the
  image, in pixels/radian. (cx,cy) are pixel coordinates corresponding to the
  projection of p = [0,0,1]. If omitted, default values are used to specify a
  normalized transverse equirectangular projection : fx=fy=1.0 and cx=cy=0.0.
  This produces q = (lat,lon)

- get_gradients: optional boolean, defaults to False. This affects what we
  return (see below)

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing arrays,
  specify them with the 'out' kwarg. If not get_gradients: 'out' is the one
  numpy array we will write into. Else: 'out' is a tuple of all the output numpy
  arrays. If 'out' is given, we return the same arrays passed in. This is the
  standard behavior provided by numpysane_pywrap.

RETURNED VALUE

if not get_gradients: we return an (...,2) array of projected transverse
equirectangular coordinates

if get_gradients: we return a tuple:

  - (...,2) array of projected transverse equirectangular coordinates
  - (...,2,3) array of the gradients of the transverse equirectangular
    coordinates in respect to the input 3D point positions

    '''

    # Internal function must have a different argument order so
    # that all the broadcasting stuff is in the leading arguments
    if not get_gradients:
        return mrcal._mrcal_npsp._project_latlon(points, fxycxy,
                                                 out=out)
    return mrcal._mrcal_npsp._project_latlon_withgrad(points, fxycxy,
                                                      out=out)


def unproject_latlon(points,
                     fxycxy = np.array((1.0, 1.0, 0.0, 0.0), dtype=float),
                     *,
                     get_gradients = False,
                     out           = None):
    r'''Unprojects 2D pixel coordinates using a transverse equirectangular projection

SYNOPSIS

    # points is a (N,2) array of imager points
    v = mrcal.unproject_latlon( points, fxycxy )

    # v is now a (N,3) array of observation directions in the camera coordinate
    # system. v are normalized

This is a special case of mrcal.unproject(). Useful not for representing lenses,
but for performing stereo rectification. Lenses do not follow this model. See
the lensmodel documentation for details:

https://mrcal.secretsauce.net/lensmodels.html#lensmodel-latlon

Given a (N,2) array of transverse equirectangular coordinates and the parameters
fxycxy, this function computes the inverse projection, optionally with
gradients.

The vectors returned by this function are normalized.

ARGUMENTS

- points: array of dims (...,2); the transverse equirectangular coordinates
  we're unprojecting. This supports broadcasting fully, and any leading
  dimensions are allowed, including none

- fxycxy: optional intrinsics core. This is a shape (4,) array (fx,fy,cx,cy),
  with all elements given in units of pixels. fx and fy are the horizontal and
  vertical focal lengths, respectively. (cx,cy) are pixel coordinates
  corresponding to the projection of p = [0,0,1]. If omitted, default values are
  used to specify a normalized transverse equirectangular projection : fx=fy=1.0
  and cx=cy=0.0. This produces q = (lat,lon)

- get_gradients: optional boolean, defaults to False. This affects what we
  return (see below)

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing arrays,
  specify them with the 'out' kwarg. If not get_gradients: 'out' is the one
  numpy array we will write into. Else: 'out' is a tuple of all the output numpy
  arrays. If 'out' is given, we return the same arrays passed in. This is the
  standard behavior provided by numpysane_pywrap.

RETURNED VALUE

if not get_gradients: we return an (...,3) array of unprojected observation
vectors. These are normalized.

if get_gradients: we return a tuple:

  - (...,3) array of unprojected observation vectors. These are normalized.
  - (...,3,2) array of the gradients of the observation vectors in respect to
    the input 2D transverse equirectangular coordinates

    '''
    if not get_gradients:
        return mrcal._mrcal_npsp._unproject_latlon(points, fxycxy,
                                                   out=out)
    return mrcal._mrcal_npsp._unproject_latlon_withgrad(points, fxycxy,
                                                        out=out)



#!/usr/bin/python3

r'''A simple camera-IMU alignment method

This routine takes measurements from rigid rig containing an IMU and some
cameras, and computes their relative geometry. The cameras are assumed to have
pre-calibrated intrinsics, which are currently trusted 100% by this tool;
weighing by uncertainty would be an improvement. The IMU is assumed to be
self-consistent, and we only use the measured gravity vector. This could be
easily extended into a more complex routine that computes the IMU biases and
such. We compute the relative translation and orientation of all the cameras,
and only an orientation from the IMU.

To gather the data, we place a stationary chessboard somewhere in the scene.
Then we collect a small number of datasets by placing the camera+IMU rig in
various orientations, where some set of cameras can observes the chessboard. We
gather chessboard images and and gravity measurements from the IMU. This allows
us to compute a set of poses to make all the measurements self-consistent.

In the sample dataset we have a set of files:

- HH-MM-SS-cami-corners.vnl: these contain the chessboard corner detections at
  time HH-MM-SS for camera i
- HH-MM-SS-gravity.vnl: these contain the IMU gravity reading at time HH-MM-SS

The HH-MM-SS timestamp is simultaneous for sets of corner observations and
gravity vectors. This timestamp is how we associate the measurements. In the
sample dataset we have:

  $ head 19-05-28-cam1-corners.vnl

  # filename x y level
  frame0000470-topic01.png 1498.253008 1150.991979 0
  frame0000470-topic01.png 1449.015013 1147.971314 0
  frame0000470-topic01.png 1399.162738 1144.527778 0
  frame0000470-topic01.png 1349.262867 1140.191815 0
  frame0000470-topic01.png 1299.681151 1134.854061 0
  frame0000470-topic01.png 1250.829060 1128.855706 0
  frame0000470-topic01.png 1202.280117 1122.228303 0
  frame0000470-topic01.png 1154.992317 1114.797872 0
  frame0000470-topic01.png 1108.771253 1107.425066 0


  $ head 19-05-28-gravity.vnl
  # acc0 acc1 acc2
  -0.8886457681655884 -7.7616963386535645 -5.7020673751831055

Here I'm optimizing for ALL the geometry, but I could assume the camera-camera
transforms are already available, to simplify

This is all a bit simplistic and could be improved, but does work quite well.

Reference coordinate system:
- camera0

Unknowns:
- rt_cam_cam0[i]   for each camera i (6*(Ncameras-1) DOF)
- rt_cam0_board[j] at each time j    (6*Nsnapshots   DOF)
- r_imu_cam0                         (3 DOF)
- g_board                            (2 DOF)

Measurements:
- q_board[i,j]
- g_imu[j]

Cost function for each i,j:
- Reprojection error q_board[i,j] - project(rt_cam_cam0[i] rt_cam0_board[j] p_board)
- r_imu_cam0 r_cam0_board[j] g_board - g_imu[j]

'''

import sys
import numpy as np
import numpysane as nps
import gnuplotlib as gp
import mrcal
import glob
import vnlog
import os
import re
import scipy




# Write the results here
Dout = '/tmp'
Din  = '/tmp/2026-01--calibrate-camera-imu-sample/',

cameras = ['cam0', 'cam1',]
models  = [mrcal.cameramodel(f"{Din}/{cam}.cameramodel") \
          for cam in cameras]

# The data is in vnl tables with the given columns. Each table is in a separate
# file, with the date/time ("YYYY-MM-DD-HH-MM-SS") in the filename identifying
# the instant in time when that data was captured. Sensor readings with the same
# tag can be assumed to have been gathered at the same instant in time. The INS
# data ends in "-gravity", the chessboard observation files have the camera name
# at the end of the filename instead.
dtype_gravity = np.dtype( [('acc0 acc1 acc2', float,(3,))])
dtype_corners = np.dtype( [('x y',            float,(2,))])
tag_regex = '/([0-9][0-9]-[0-9][0-9]-[0-9][0-9])'
glob_files_gravity = f'{Din}/*-gravity.vnl'
glob_files_corners = [f'{Din}/*-{cam}-corners.vnl' for cam in cameras]

# chessboard
gridn          = 14
object_spacing = 0.0588

# relative scale of pixels errors to IMU axis errors. These should match the
# uncertainty of each respective measurement
imu_err_scale__pixels_per_m_s2 = 0.3/0.05



files_gravity         = sorted(glob.glob(glob_files_gravity))
files_corners_allcams = [sorted(glob.glob(g)) for g in glob_files_corners]


def tag_from_file(f):
    m = re.search(tag_regex,f)
    return m.group(1)

_isnapshot_from_tag = dict()
for i,f in enumerate(files_gravity):
    _isnapshot_from_tag[ tag_from_file(f) ] = i

def isnapshot_from_file(f):
    return _isnapshot_from_tag[tag_from_file(f)]


Nsnapshots = len(files_gravity)
g_imu = np.zeros( (Nsnapshots,3), dtype=float)
for f in files_gravity:
    g = vnlog.slurp(f, dtype=dtype_gravity)['acc0 acc1 acc2']
    g_imu[isnapshot_from_file(f)] = g
g_imu_mag = nps.mag(g_imu)
gunit_imu = g_imu / nps.dummy( g_imu_mag, -1)

Ncameras = len(cameras)
icam_from_cam = dict()
for i,c in enumerate(cameras):
    icam_from_cam[c] = i


Nobservations = sum(len(files_corners) for files_corners in files_corners_allcams)

# qx,qy will be filled in later; weight=1 is hard-coded here now
observations_qxqyw   = np.ones ( (Nobservations,gridn,gridn,3), dtype=float)
indices_frame_camera = np.zeros( (Nobservations,2), dtype=np.int32 )
i = 0
for icam in range(Ncameras):
    for f in files_corners_allcams[icam]:
        isnapshot = isnapshot_from_file(f)
        observations_qxqyw[i,...,:2] = vnlog.slurp(f, dtype=dtype_corners)['x y'].reshape((gridn,gridn,2))
        indices_frame_camera[i,0] = isnapshot
        indices_frame_camera[i,1] = icam
        i += 1


######## Seed
# I need to estimate each of the state:
# - rt_cam_cam0[i]   for each camera i (6*(Ncameras-1) DOF)
# - rt_cam0_board[j] at each time j    (6*Nsnapshots   DOF)
# - r_imu_cam0                         (3 DOF)
# - g_board                            (2 DOF)

# For each observation, estimate the cam-board transform. This assumes a nearby
# pinhole model, so it's a rough estimate. I take arbitrary subsets of these to
# get rt_cam_cam0 and rt_cam0_board
Rt_cam_board_all = \
    mrcal.estimate_monocular_calobject_poses_Rt_tocam( indices_frame_camera,
                                                       observations_qxqyw,
                                                       object_spacing,
                                                       models)

rt_cam_cam0   = np.zeros( (Ncameras-1, 6), dtype=float)
rt_cam0_board = np.zeros( (Nsnapshots,6),  dtype=float)
Nrt_cam_cam0_have   = 0
Nrt_cam0_board_have = 0

# I go through to accumulate my seed.
#### WARNING: THIS DOES NOT WORK IN GENERAL
#### HERE I'M ASSUMING THAT CAMERA0 IS IN EVERY SNAPSHOT
for iobservation in range(Nobservations):
    Rt_cam_board   = Rt_cam_board_all[iobservation]
    isnapshot,icam = indices_frame_camera[iobservation]
    if icam==0 and not np.any(rt_cam0_board[isnapshot]):
        rt_cam0_board[isnapshot] = mrcal.rt_from_Rt(Rt_cam_board)
        Nrt_cam0_board_have += 1
        if Nrt_cam0_board_have == len(rt_cam0_board):
            break
if Nrt_cam0_board_have != len(rt_cam0_board):
    raise Exception("ERROR: did not init all of rt_cam0_board")

for iobservation in range(Nobservations):
    Rt_cam_board   = Rt_cam_board_all[iobservation]
    isnapshot,icam = indices_frame_camera[iobservation]
    if icam!=0 and not np.any(rt_cam_cam0[icam-1]):
        rt_cam_cam0[icam-1] = \
            mrcal.compose_rt(mrcal.rt_from_Rt(Rt_cam_board),
                             rt_cam0_board[isnapshot],
                             inverted1 = True)
        Nrt_cam_cam0_have += 1
        if Nrt_cam_cam0_have == len(rt_cam_cam0):
            break
if Nrt_cam_cam0_have != len(rt_cam_cam0):
    raise Exception("ERROR: did not init all of rt_cam_cam0")

######## Solve

def unpack_state(b, have_imu = False):
    Nstate_cameras = (Ncameras-1 + Nsnapshots) * 6
    bcameras = b[:Nstate_cameras].reshape( (Ncameras-1 + Nsnapshots, 6) )
    rt_cam_cam0   = bcameras[:Ncameras-1,:]
    rt_cam0_board = bcameras[Ncameras-1:,:]

    if not have_imu:
        return rt_cam_cam0,rt_cam0_board

    bimu = b[Nstate_cameras:]
    r_imu_cam0 = bimu[:3]

    a,b = bimu[3:]
    gunit_board = np.array(( np.sin(a),
                             np.cos(a)*np.cos(b),
                             np.cos(a)*np.sin(b),))
    return rt_cam_cam0,rt_cam0_board,r_imu_cam0,gunit_board

def pack_state(rt_cam_cam0, rt_cam0_board,
               r_imu_cam0 = None,
               gunit_board = None):
    if r_imu_cam0 is None:
        return \
            nps.cat( *rt_cam_cam0, *rt_cam0_board ).ravel()

    # have_imu
    a = np.arcsin( gunit_board[0])
    c = np.cos(a)
    if np.abs(c) < 1e-8:
        b = 0
    else:
        b = np.arctan2(gunit_board[2]/c,
                       gunit_board[1]/c,)
    return \
        nps.glue( rt_cam_cam0.ravel(),
                  rt_cam0_board.ravel(),
                  r_imu_cam0.ravel(),
                  a,
                  b,
                  axis = -1)

def optimizer_callback_camera(b, have_imu = False, debug = False):
    if not have_imu:
        rt_cam_cam0,rt_cam0_board = unpack_state(b, have_imu = have_imu)
    else:
        rt_cam_cam0,rt_cam0_board,r_imu_cam0,gunit_board = unpack_state(b, have_imu = have_imu)

    x_corners = np.zeros((Nobservations,gridn,gridn,2), dtype=float)

    for iobservation in range(Nobservations):
        isnapshot,icam = indices_frame_camera[iobservation]
        q_observed = observations_qxqyw[iobservation, ..., :2]

        pcam = mrcal.transform_point_rt( rt_cam0_board[isnapshot],
                                         calobject )
        if icam > 0:
            pcam = mrcal.transform_point_rt( rt_cam_cam0[icam-1],
                                             pcam )
        q = mrcal.project(pcam, *models[icam].intrinsics())
        x_corners[iobservation,...] = q - q_observed

    if debug:
        print(f"Camera solve callback RMS error: {np.sqrt(np.mean(nps.norm2(x_corners).ravel()))}")

    if not have_imu:
        return x_corners.ravel()

    gunit_cam0      = mrcal.rotate_point_r(rt_cam0_board[:,:3], gunit_board)
    gunit_imu_solve = mrcal.rotate_point_r(r_imu_cam0, gunit_cam0)
    g_imu_solve = gunit_imu_solve * 9.8
    x_imu = g_imu_solve - g_imu
    if debug:
        print(f"IMU vector error: {x_imu}m/s^2")

    return nps.glue(x_corners.ravel(),
                    x_imu.ravel() * imu_err_scale__pixels_per_m_s2,
                    axis = -1)


###### I solve the camera stuff by itself initially, since that should fit very
###### well
calobject = mrcal.ref_calibration_object(gridn, gridn,
                                         object_spacing)

# state vector at the seed
b0 = pack_state(rt_cam_cam0, rt_cam0_board)
result = scipy.optimize.least_squares(optimizer_callback_camera, b0,
                                      kwargs=dict(have_imu = False))

# if True:
#     print("At the solution:")
#     optimizer_callback_camera(result['x'], debug = True)
#     mrcal.show_geometry( nps.glue( mrcal.identity_rt(),
#                                    rt_cam_cam0,
#                                    axis = -2),
#                          wait = True)


###### Now do a bigger optimization, including the IMU stuff
rt_cam_cam0,rt_cam0_board = unpack_state(result['x'])

####### For now I assume the board is sitting roughly vertically, and seed off
####### that. THIS IS NOT TRUE IN GENERAL
gunit_board = np.array((0., -1., 0.))
gunit_cam0 = mrcal.rotate_point_r(rt_cam0_board[:,:3], gunit_board)
R_imu_cam0 = mrcal.align_procrustes_vectors_R01(gunit_imu, gunit_cam0)
r_imu_cam0 = mrcal.r_from_R(R_imu_cam0)
gunit_imu_seed = mrcal.rotate_point_R(R_imu_cam0, gunit_cam0)
th_err_imu_seed = np.arccos(nps.inner(gunit_imu_seed, gunit_imu))
if False:
    # should be "small"
    print(th_err_imu_seed)

b0 = pack_state(rt_cam_cam0, rt_cam0_board, r_imu_cam0, gunit_board)
result = scipy.optimize.least_squares(optimizer_callback_camera, b0,
                                      kwargs=dict(have_imu = True))
rt_cam_cam0,rt_cam0_board,r_imu_cam0,gunit_board = unpack_state(result['x'], have_imu = True)

rt_cam_cam0_mounted = \
    nps.glue( mrcal.identity_rt(),
              rt_cam_cam0,
              axis = -2)
rt_imu_cam0 = nps.glue( r_imu_cam0, np.zeros((3,)),
                        axis = -1)

for icam in range(Ncameras):
    cam = cameras[icam]
    rt_cam_cam0_this = rt_cam_cam0_mounted[icam]

    filename = f"{Dout}/{cam}-final.cameramodel"
    m = mrcal.cameramodel(models[icam])
    m.rt_cam_ref( rt_cam_cam0_this )
    m.write(filename)
    print(f"Wrote '{filename}'")
# And a dummy imu model; for visualization only
filename = f"{Dout}/imu-final.cameramodel"
m = mrcal.cameramodel(rt_cam_ref = rt_imu_cam0,
                      # dummy
                      intrinsics = ('LENSMODEL_PINHOLE', np.array((1,1,0,0))),
                      imagersize = (1,1) )
m.write(filename)
print(f"Wrote '{filename}'")

if True:
    print("At the solution:")
    optimizer_callback_camera(result['x'],
                              have_imu = True,
                              debug = True)

    gunit_cam0  = mrcal.rotate_point_r(r_imu_cam0,          gunit_imu,  inverted=True)
    gunit_board = mrcal.rotate_point_r(rt_cam0_board[:,:3], gunit_cam0, inverted=True)
    print(f"guinit_board=\n{gunit_board}")

    mrcal.show_geometry( nps.glue( rt_cam_cam0_mounted,
                                   rt_imu_cam0,
                                   axis = -2),
                         title = "Camera-IMU calibration result",
                         cameranames = cameras + ['imu'],
                         wait = True)


    if True:
        # plots for the documentation
        mrcal.show_geometry( nps.glue( rt_cam_cam0_mounted,
                                       rt_imu_cam0,
                                       axis = -2),
                             cameranames = cameras + ['imu'],
                             title = "Camera-IMU calibration result",
                             hardcopy="/home/dima/projects/mrcal-doc-external/figures/2026-01--calibrate-camera-imu-sample/geometry.png",
                             terminal='pngcairo size 1024,768 transparent noenhanced crop font ",12"')
        mrcal.show_geometry( nps.glue( rt_cam_cam0_mounted,
                                       rt_imu_cam0,
                                       axis = -2),
                             cameranames = cameras + ['imu'],
                             title = "Camera-IMU calibration result",
                             hardcopy="/home/dima/projects/mrcal-doc-external/figures/2026-01--calibrate-camera-imu-sample/geometry.pdf",
                             terminal='pdf size 8in,6in noenhanced solid color   font ",16"')



#!/usr/bin/env python3

'''This is a set of simple experiments to test the outlier-rejection and
sensitivity logic'''


import sys
import os

import numpy as np
import numpysane as nps
import gnuplotlib as gp


usage = "Usage: {} order Npoints noise_stdev".format(sys.argv[0])
if len(sys.argv) != 4:
    print usage
    sys.exit(1)
try:
    order       = int(  sys.argv[1]) # order. >= 1
    N           = int(  sys.argv[2]) # This many points in the dataset
    noise_stdev = float(sys.argv[3]) # The dataset is corrupted thusly
except:
    print usage
    sys.exit(1)
if order < 1 or N <= 0 or noise_stdev <= 0:
    print usage
    sys.exit(1)


Nquery = 70    # This many points for post-fit uncertainty-evaluation



reference_equation = '0.1*(x+0.2)**2. + 3.0'

def report_mismatch_relerr(a,b, what):
    relerr = np.abs(a-b) / ( (a+b)/2.)
    if relerr > 1e-6:
        print "MISMATCH for {}: relerr = {}, a = {}, b = {},".format(what,relerr,a,b);

def report_mismatch_abserr(a,b, what):
    abserr = np.abs(a-b)
    if abserr > 1e-6:
        print "MISMATCH for {}: abserr = {}, a = {}, b = {},".format(what,abserr,a,b);

def model_matrix(q, order):
    r'''Returns the model matrix S for particular domain points

    Here the "order" is the number of parameters in the fit. Thus order==2 means
    "linear" and order==3 means "quadratic""

    '''
    q = nps.atleast_dims(q,-1)
    return nps.transpose(nps.cat(*[q ** i for i in range(order)]))

def func_reference(q):
    '''Reference function: reference_equation

    Let's say I care about the range [0..1]. It gets less linear as I get
    further away from 0

    '''

    # needs to manually match 'reference_equation' above
    return 0.1 * (q+0.2)*(q+0.2) + 3.0

def func_hypothesis(q, b):
    '''Hypothesis based on parameters

    '''
    S = model_matrix(q, len(b))
    return nps.matmult(b, nps.transpose(S))

def compute_outliernesses(J, x, jq, k_dima, k_cook):
    '''Computes all the outlierness/Cook's D metrics

    I have 8 things I can compute coming from 3 yes/no choices. These are all
    very similar, with two pairs actually coming out identical. I choose:

    - Are we detecting outliers, or looking at effects of a new query point?
    - Dima's outlierness factor or Cook's D
    - Look ONLY at the effect on the other variables, or on the other variables
      AND self?

    If we're detecting outliers, we REMOVE measurements from the dataset, and
    see what happens to the fit. If we're looking at effects of a new query
    point, we see what happend if we ADD measurements

    Dima's outlierness factor metric looks at what happens to the cost function
    E = norm2(x). Specifically I look at

      (norm2(x_before) - norm(x_after))/Nmeasurements

    Cook's D instead looks at

      (norm2(x_before - x_after)) * k

    for some constant k.

    Finally, we can decide whether to include the effects on the measurements
    we're adding/removing, or not.

    Note that here I only look at adding/removing SCALAR measurements


    =============




    This is similar-to, but not exactly-the-same-as Cook's D. I assume the least
    squares fit optimizes a cost function E = norm2(x). The outlierness factor I
    return is

      f = 1/Nmeasurements (E(outliers and inliers) - E(inliers only))

    For a scalar measurement, this solves to

      k = xo^2 / Nmeasurements
      B = 1.0/(jt inv(JtJ) j - 1)
      f = -k * B

    (see the comment in dogleg_getOutliernessFactors() for a description)

    Note that my metric is proportional to norm2(x_io) - norm2(x_i). This is NOT
    the same as Cook's distance, which is proportional to norm2(x_io - x_i).
    It's not yet obvious to me which is better





    There're several slightly-different definitions of Cook's D and of a
    rule-of-thumb threshold floating around on the internet. Wikipedia says:

      D = norm2(x_io - x_i)^2 / (Nstate * norm2(x_io)/(Nmeasurements - Nstate))
      D_threshold = 1

    An article https://www.nature.com/articles/nmeth.3812 says

      D = norm2(x_io - x_i)^2 / ((Nstate+1) * norm2(x_io)/(Nmeasurements - Nstate -1))
      D_threshold = 4/Nmeasurements

    Here I use the second definition. That definition expands to

      k = xo^2 / ((Nstate+1) * norm2(x_io)/(Nmeasurements - Nstate -1))
      B = 1.0/(jt inv(JtJ) j - 1)
      f = k * (B + B*B)

    '''

    Nmeasurements,Nstate = J.shape

    # The A values for each measurement
    Aoutliers = nps.inner(J, nps.transpose(np.linalg.pinv(J)))
    Aquery    = nps.inner(jq, nps.transpose(np.linalg.solve(nps.matmult(nps.transpose(J),J), nps.transpose(jq))))

    def dima():

        k = k_dima
        k = 1

        # Here the metrics are linear, so self + others = self_others
        def outliers():
            B = 1.0 / (Aoutliers - 1.0)
            return dict( self        = k * x*x,
                         others      = k * x*x*(-B-1),
                         self_others = k * x*x*(-B  ))
        def query():
            B = 1.0 / (Aquery + 1.0)
            return dict( self        = k * (  B*B),
                         others      = k * (B-B*B),
                         self_others = k * (B))
        return dict(outliers = outliers(),
                    query    = query())
    def cook():

        k = k_cook
        k = 1

        # Here the metrics maybe aren't linear (I need to think about it), so
        # maybe self + others != self_others. I thus am not returning the "self"
        # metric
        def outliers():
            B = 1.0 / (Aoutliers - 1.0)
            return dict( self_others = k * x*x*(B+B*B  ) ,
                         others      = k * x*x*(-B-1))
        def query():
            B = 1.0 / (Aquery + 1.0)
            return dict( self_others = k * (1-B) ,
                         others      = k * (B-B*B))
        return dict(outliers = outliers(),
                    query    = query())


    return dict(cook = cook(),
                dima = dima())

def outlierness_test(J, x, f, outlierness, k_dima, k_cook, i=0):
    r'''Test the computation of outlierness

    I have an analytical expression for this computed in
    compute_outliernesses(). This explicitly computes the quantity represented
    by compute_outliernesses() to make sure that that analytical expression is
    correct

    '''

    # I reoptimize without measurement i
    E0 = nps.inner(x,x)

    J1 = nps.glue(J[:i,:], J[(i+1):,:], axis=-2)
    f1 = nps.glue(f[:i  ], f[(i+1):  ], axis=-1)
    b1 = nps.matmult( f1, nps.transpose(np.linalg.pinv(J1)))
    x1 = nps.matmult(b1, nps.transpose(J1)) - f1
    E1 = nps.inner(x1,x1)

    report_mismatch_relerr( (E0-E1) * k_dima,
                            outlierness['self_others'][i],
                            "self_others outlierness computed analytically, explicitly")
    report_mismatch_relerr( (E0-x[i]*x[i] - E1) * k_dima,
                            outlierness['others'][i],
                            "others outlierness computed analytically, explicitly")

def CooksD_test(J, x, f, CooksD, k_dima, k_cook, i=0):
    r'''Test the computation of Cook's D

    I have an analytical expression for this computed in
    compute_outliernesses(). This explicitly computes the quantity represented
    by compute_outliernesses() to make sure that that analytical expression is
    correct

    '''

    # I reoptimize without measurement i
    Nmeasurements,Nstate = J.shape

    J1 = nps.glue(J[:i,:], J[(i+1):,:], axis=-2)
    f1 = nps.glue(f[:i  ], f[(i+1):  ], axis=-1)
    b1 = nps.matmult( f1, nps.transpose(np.linalg.pinv(J1)))
    x1 = nps.matmult(b1, nps.transpose(J)) - f

    dx = x1-x

    report_mismatch_relerr( nps.inner(dx,dx) * k_cook,
                            CooksD['self_others'][i],
                            "self_others CooksD computed analytically, explicitly")
    report_mismatch_relerr( (nps.inner(dx,dx) - dx[i]*dx[i]) * k_cook,
                            CooksD['others'][i],
                            "others CooksD computed analytically, explicitly")

def outlierness_query_test(J,b,x, f, query,fquery_ref, outlierness_nox, k_dima, k_cook, i=0):
    r'''Test the concept of outlierness for querying hypothetical data

    fquery_test = f(q) isn't true here. If it WERE true, the x of the query
    point would be 0 (we fit the model exactly), so the outlierness factor would
    be 0 also

    '''

    # current solve
    E0 = nps.inner(x,x)

    query      = query     [i]
    fquery_ref = fquery_ref[i]


    # I add a new point, and reoptimize
    fquery = func_hypothesis(query,b)
    xquery = fquery - fquery_ref
    jquery = model_matrix(query, len(b))

    J1 = nps.glue(J, jquery,     axis=-2)
    f1 = nps.glue(f, fquery_ref, axis=-1)
    b1 = nps.matmult( f1, nps.transpose(np.linalg.pinv(J1)))
    x1 = nps.matmult(b1, nps.transpose(J1)) - f1
    E1 = nps.inner(x1,x1)

    report_mismatch_relerr( (x1[-1]*x1[-1]) * k_dima,
                            outlierness_nox['self'][i]*xquery*xquery,
                            "self query-outlierness computed analytically, explicitly")
    report_mismatch_relerr( (E1-x1[-1]*x1[-1] - E0) * k_dima,
                            outlierness_nox['others'][i]*xquery*xquery,
                            "others query-outlierness computed analytically, explicitly")
    report_mismatch_relerr( (E1 - E0) * k_dima,
                            outlierness_nox['self_others'][i]*xquery*xquery,
                            "self_others query-outlierness computed analytically, explicitly")

def CooksD_query_test(J,b,x, f, query,fquery_ref, CooksD_nox, k_dima, k_cook, i=0):
    r'''Test the concept of CooksD for querying hypothetical data

    fquery_test = f(q) isn't true here. If it WERE true, the x of the query
    point would be 0 (we fit the model exactly), so the outlierness factor would
    be 0 also

    '''

    # current solve
    Nmeasurements,Nstate = J.shape

    query      = query     [i]
    fquery_ref = fquery_ref[i]

    # I add a new point, and reoptimize
    fquery = func_hypothesis(query,b)
    xquery = fquery - fquery_ref
    jquery = model_matrix(query, len(b))

    J1 = nps.glue(J, jquery,     axis=-2)
    f1 = nps.glue(f, fquery_ref, axis=-1)
    b1 = nps.matmult( f1, nps.transpose(np.linalg.pinv(J1)))
    x1 = nps.matmult(b1, nps.transpose(J1)) - f1

    dx = x1[:-1] - x

    dx_both = x1 - nps.glue(x,xquery, axis=-1)

    report_mismatch_relerr( nps.inner(dx_both,dx_both)*k_cook,
                            CooksD_nox['self_others'][i]*xquery*xquery,
                            "self_others query-CooksD computed analytically, explicitly")
    report_mismatch_relerr( nps.inner(dx,dx)*k_cook,
                            CooksD_nox['others'][i]*xquery*xquery,
                            "others query-CooksD computed analytically, explicitly")



def Var_df(J, squery, stdev):
    r'''Propagates noise in input to noise in f

    noise in input -> noise in params -> noise in f

    db ~ M dm where M = inv(JtJ)Jt

    df = df/db db

    df/db = squery
    Var(dm) = stdev^2 I ->

    Var(df) = stdev^2 squery inv(JtJ) Jt J inv(JtJ) squeryt =
            = stdev^2 squery inv(JtJ) squeryt

    This function broadcasts over squery
    '''
    return \
        nps.inner(squery,
                  nps.transpose(np.linalg.solve(nps.matmult(nps.transpose(J),J),
                                                nps.transpose(squery)))) *stdev*stdev

def histogram(x, **kwargs):
    h,edges = np.histogram(x, bins=20, **kwargs)
    centers = (edges[1:] + edges[0:-1])/2
    return h,centers,edges[1]-edges[0]


def generate_dataset(N, noise_stdev):
    q      = np.random.rand(N)
    fref   = func_reference(q)
    fnoise = np.random.randn(N) * noise_stdev
    f      = fref + fnoise
    return q,f


def fit(q, f, order):
    S = model_matrix(q, order)
    J = S
    b = nps.matmult( f, nps.transpose(np.linalg.pinv(S)))
    x = func_hypothesis(q,b) - f
    return b,J,x


def test_order(q,f, query, order):

    # I look at linear and quadratic models: a0 + a1 q + a2 q^2, with a2=0 for the
    # linear case. I use plain least squares. The parameter vector is [a0 a1 a2]t. S
    # = [1 q q^2], so the measurement vector x = S b - f. E = norm2(x). J = dx/db =
    # S.
    #
    # Note the problem "order" is the number of parameters, so a linear model has
    # order==2
    b,J,x = fit(q,f,order)

    Nmeasurements,Nstate = J.shape
    k_dima = 1.0/Nmeasurements
    k_cook = 1.0/((Nstate + 1.0) * nps.inner(x,x)/(Nmeasurements - Nstate - 1.0))

    report_mismatch_abserr(np.linalg.norm(nps.matmult(x,J)), 0, "Jtx")

    squery = model_matrix(query, order)
    fquery = func_hypothesis(query, b)
    metrics = compute_outliernesses(J,x, squery, k_dima, k_cook)

    outlierness_test(J, x, f, metrics['dima']['outliers'], k_dima, k_cook, i=10)
    CooksD_test     (J, x, f, metrics['cook']['outliers'], k_dima, k_cook, i=10)
    outlierness_query_test(J,b,x,f, query, fquery + 1.2e-3, metrics['dima']['query'], k_dima, k_cook, i=10 )
    CooksD_query_test     (J,b,x,f, query, fquery + 1.2e-3, metrics['cook']['query'], k_dima, k_cook, i=10 )

    Vquery = Var_df(J, squery, noise_stdev)
    return \
        dict( b       = b,
              J       = J,
              x       = x,
              Vquery  = Vquery,
              squery  = squery,
              fquery  = fquery,
              metrics = metrics,
              k_dima  = k_dima,
              k_cook  = k_cook )


q,f   = generate_dataset(N, noise_stdev)
query = np.linspace(-1,2, Nquery)

stats = test_order(q,f, query, order)


def show_outlierness(order, N, q, f, query, cooks_threshold, **stats):

    p = gp.gnuplotlib(equation='1.0 title "Threshold"',
                      title   = "Outlierness with order={} Npoints={} stdev={}".format(order, N, noise_stdev))
    p.plot( (q, stats['metrics']['dima']['outliers']['self_others']/stats['dimas_threshold'],
             dict(legend="Dima's self+others outlierness / threshold",
                  _with='points')),
            (q, stats['metrics']['dima']['outliers']['others']/stats['dimas_threshold'],
             dict(legend="Dima's others-ONLY outlierness / threshold",
                  _with='points')),
            (q, stats['metrics']['cook']['outliers']['self_others']/cooks_threshold,
             dict(legend="Cook's self+others outlierness / threshold",
                  _with='points')),
            (q, stats['metrics']['cook']['outliers']['others']/cooks_threshold,
             dict(legend="Cook's others-ONLY outlierness / threshold",
                  _with='points')))
    return p

def show_uncertainty(order, N, q, f, query, cooks_threshold, **stats):

    coeffs = stats['b']
    fitted_equation = '+'.join(['{} * x**{}'.format(coeffs[i], i) for i in range(len(coeffs))])

    p = gp.gnuplotlib(equation='({})-({}) title "Fit error off ground truth; y2 axis +- noise stdev" axis x1y2'.format(reference_equation,fitted_equation),
                      title   = "Uncertainty with order={} Npoints={} stdev={}".format(order, N, noise_stdev),
                      ymin=0,
                      y2range = (-noise_stdev, noise_stdev),
                      _set = 'y2tics'
)
    # p.plot(
    #     (query, np.sqrt(stats['Vquery]),
    #      dict(legend='expectederr (y2)', _with='lines', y2=1)),

    #          (query, stats['metrics']['dima']['query']['self_others']*noise_stdev*noise_stdev / stats['dimas_threshold'],
    #           dict(legend="Dima's self+others query / threshold",
    #                _with='linespoints')),
    #     (query, stats['metrics']['dima']['query']['others']*noise_stdev*noise_stdev / stats['dimas_threshold'],
    #      dict(legend="Dima's others-ONLY query / threshold",
    #           _with='linespoints')),

    #          (query, stats['metrics']['cook']['query']['self_others']*noise_stdev*noise_stdev / cooks_threshold,
    #           dict(legend="Cook's self+others query / threshold",
    #                _with='linespoints')),
    #     (query, stats['metrics']['cook']['query']['others']*noise_stdev*noise_stdev / cooks_threshold,
    #      dict(legend="Cook's others-ONLY query / threshold",
    #           _with='linespoints')))

    p.plot(

        # (query, np.sqrt(Vquery),
        #  dict(legend='Expected error due to input noise (y2)', _with='lines', y2=1)),

        (query, np.sqrt(stats['metrics']['dima']['query']['self'])*noise_stdev,
         dict(legend="Dima's self-ONLY; 1 point",
              _with='linespoints')),

        (query, np.sqrt(stats['metrics']['dima']['query']['others'])*noise_stdev,
         dict(legend="Dima's others-ONLY ALL {} points".format(Nquery),
              _with='linespoints')),

        # (query, np.sqrt(stats['metrics']['dima']['query']['others'])*noise_stdev * Nquery,
        #  dict(legend="Dima's others-ONLY 1 point average",
        #       _with='linespoints')),


          )

    return p

def show_fit        (order, N, q, f, query, cooks_threshold, **stats):
    p = gp.gnuplotlib(equation='{} with lines title "reference"'.format(reference_equation),
                      xrange=[-1,2],
                      title = "Fit with order={} Npoints={} stdev={}".format(order, N, noise_stdev))
    p.plot((q,     f,                                                                       dict(legend = 'input data', _with='points')),
           (query, stats['fquery'] + np.sqrt(stats['Vquery'])*np.array(((1,),(0,),(-1,),)), dict(legend = 'stdev_f',    _with='lines')))
    return p

def show_distribution(outlierness):
    h,c,w = histogram(outlierness)
    raise Exception("not done")
    #gp.plot()



# This is hoaky, but reasonable, I think. Some of the unscaled metrics are
# identical between mine and Cook's expressions. So I scale Cook's 4/N threshold
# to apply to me. Works ok.
cooks_threshold  = 4.0 / N
stats['dimas_threshold'] = cooks_threshold / stats['k_cook'] * stats['k_dima']

# These all show interesting things; turn one of them on
plots = [ show_outlierness (order, N, q, f, query, cooks_threshold, **stats),
          show_fit         (order, N, q, f, query, cooks_threshold, **stats),
          show_uncertainty (order, N, q, f, query, cooks_threshold, **stats)
        ]

for p in plots:
    if os.fork() == 0:
        p.wait()
        sys.exit()
os.wait()


# Conclusions:
#
# - Cook's 4/N threshold looks reasonable.
#
# - For detecting outliers my self+others metric is way too outlier-happy. It
#   makes most of the linear fit points look outliery
#
# - For uncertainty, Dima's self+others is the only query metric that's maybe
#   usable. It qualitatively a negated and shifted Cook's self+others



#!/usr/bin/env python3

r'''Basic structure-from-motion test

I observe, with noise, a number of points from various angles with a single
camera, and I make sure that I can accurately compute the locations of the
camera.

I simulate a horizontal plane with many features on it, with the forward-looking
camera moving around just above this horizontal plane (like a car driving around
a flat-ish area)

'''

import sys
import numpy as np
import numpysane as nps
import os

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils

import numpy.random
np.random.seed(0)


############# Set up my world, and compute all the perfect positions, pixel
############# observations of everything
(W,H) = (4000,2200)
m = mrcal.cameramodel( intrinsics = ('LENSMODEL_PINHOLE',
                                     np.array((600., 600., (W-1)/2, (H-1)/2))),
                       imagersize = (W,H) )

# Camera0 is fixed: it's the reference coordinate system in the solve. The
# points I'm observing lie on a perfect horizontal plane (y = constant) sitting
# a bit below the cameras. The points span a forward arc at different ranges
ranges = np.array(( 1000,
                    500,
                    200,
                    100,
                    50,
                    30,
                    20,
                    15,
                    8,
                    5,
                    4,
                    3,
                   ))

points_fov_deg    = 160.
Npoints_per_range = 50

# The camera starts out at the center, and moves steadily to the right along the
# x axis, with some smaller nonlinear y motion as well. I want to avoid
# collinear camera locations (this produces singular solves). Orientation is
# always straight ahead (pointed along the z axis) and there's no front/back motion
Ncamera_poses = 10
step          = 1.0



th = np.linspace(-points_fov_deg/2.*np.pi/180.,
                  points_fov_deg/2.*np.pi/180.,
                 Npoints_per_range)
cth = np.cos(th)
sth = np.sin(th)

# In the ref coordinate system (locked to camera0)
v = nps.glue( nps.transpose(sth),
              np.zeros(cth.shape + (1,)),
              nps.transpose(cth),
              axis = -1)

# In the ref coordinate system (locked to camera0)
# Baseline is flat, below the cameras. Curved left/right. Rising as we move away
# from the cameras
# shape (Npoints, 3)
points_true = \
    nps.clump( v * nps.mv(ranges,  -1,-3),
               n = 2 ) + \
    np.array((0, 1.5, 0))
points_true[:,1] += (points_true[:,0]/ranges[0]*15.)**2.
points_true[:,1] -= (points_true[:,2]/ranges[0]*200.)

Npoints = points_true.shape[0]

if False:
    import gnuplotlib as gp
    gp.plot(points_true,
            _3d=1,
            xlabel='x',
            ylabel='y',
            zlabel='z',
            _with='points',
            square=1,
            tuplesize=-3,
            wait=1)
    sys.exit()

x_ref_cam_true  = np.arange(Ncamera_poses) * step
y_ref_cam_true  = np.zeros((Ncamera_poses,),)
y_ref_cam_true  = (np.arange(Ncamera_poses) - (Ncamera_poses-1.)/2.) ** 2. / 100.
y_ref_cam_true -= y_ref_cam_true[0]

# shape (Ncamera_poses,6)
rt_ref_cam_true = nps.glue( np.zeros((Ncamera_poses,3)), # r
                            nps.transpose(x_ref_cam_true),
                            nps.transpose(y_ref_cam_true),
                            np.zeros((Ncamera_poses,1)), # z
                            axis = -1 )
rt_cam_ref_true = mrcal.invert_rt(rt_ref_cam_true)

if False:
    mrcal.show_geometry(rt_cam_ref_true,
                        #points = points_true
                        )

# I project all the points into all the cameras. Anything that's in view, I keep

# shape (Npoints, Ncamera_poses, 3)
pcam_true = mrcal.transform_point_rt(rt_cam_ref_true,
                                     nps.mv(points_true, -2, -3))

# shape (Npoints, Ncamera_poses, 2)
qcam_true = mrcal.project(pcam_true, *m.intrinsics())

# ALL the indices. I'm about to cut these down to the visible ones
# shape (Npoints, Ncamera_poses)
indices_cam, indices_point = \
    np.meshgrid(np.arange(Ncamera_poses),
                np.arange(Npoints))

valid_observation_index = \
    (qcam_true[...,0] >= 0  ) * \
    (qcam_true[...,0] <= W-1) * \
    (qcam_true[...,1] >= 0  ) * \
    (qcam_true[...,1] <= H-1) * \
    (pcam_true[...,2] >  0)

indices_point_camintrinsics_camextrinsics = \
    nps.glue( nps.transpose(indices_point[valid_observation_index]),
              np.zeros((np.count_nonzero(valid_observation_index),1)),
              nps.transpose(indices_cam  [valid_observation_index]) - 1,
              axis = -1 ).astype(np.int32)

observations_true = qcam_true[valid_observation_index]

# Any point observed by a single camera is thrown out
#
# I have a sequence of observed-point indices. Most point are observed by
# multiple cameras, so the indices will appear more than once. But some indices
# will appear just once, and I want to throw those away. How do I find these?
#
# 1. I compute the diff: nonzero entries signify transitions between different
#    points. Single points will have consecutive transitions. So...
# 2. diff(diff(ipoint)) == 0 AND both sides are a transition signifies single
#    points
ipoint = indices_point_camintrinsics_camextrinsics[:,0]
d = nps.glue(True,
             np.diff(ipoint).astype(bool),
             True,
             axis=-1)
ipoint_not_single_mask = np.diff(d) + np.logical_not(d[:-1])

indices_point_camintrinsics_camextrinsics = \
    indices_point_camintrinsics_camextrinsics[ipoint_not_single_mask]
observations_true = \
    observations_true[ipoint_not_single_mask]

# I just threw away points with a single observation, which created some gaps in
# the point arrays. I reorder the point array and the array referencing it so
# that the remaining points are indexed with sequential integers, starting at 0

Npoints_old = Npoints

ipoint = indices_point_camintrinsics_camextrinsics[:,0]
ipoint_unique = np.unique(ipoint)
new_point_delta = np.zeros((Npoints_old,), dtype=int)
new_point_delta[ipoint_unique] = 1

# shape (Npoints_old,)
# For each i=ipoint_old I have ipoint_new = index_ipoint_new[i]
index_ipoint_new = np.cumsum(new_point_delta) - 1
# shape (Npoints_new,)
# For each i=ipoint_new I have ipoint_old = index_ipoint_old[i]
index_ipoint_old = ipoint_unique

indices_point_camintrinsics_camextrinsics[:,0] = \
    index_ipoint_new[indices_point_camintrinsics_camextrinsics[:,0]]
points_true = points_true[index_ipoint_old]

Npoints = len(ipoint_unique)

r_cam_ref_noise_rad = 0.1
t_cam_ref_noise_m   = 0.5
r_cam_ref_noise = \
    ((np.random.random_sample(rt_cam_ref_true[:,:3].shape) * 2) - 1) * r_cam_ref_noise_rad
t_cam_ref_noise = \
    ((np.random.random_sample(rt_cam_ref_true[:,3:].shape) * 2) - 1) * t_cam_ref_noise_m
# first camera is at the origin
r_cam_ref_noise[0] *= 0
t_cam_ref_noise[0] *= 0

rt_cam_ref_noisy = rt_cam_ref_true.copy()
rt_cam_ref_noisy[:,:3] += r_cam_ref_noise
rt_cam_ref_noisy[:,3:] += t_cam_ref_noise

observations_noise_pixels = 2
observations_noise = \
    ((np.random.random_sample(observations_true.shape) * 2) - 1) * observations_noise_pixels
observations_noisy = observations_true + observations_noise


observations = np.array(observations_noisy)
rt_cam_ref   = np.array(rt_cam_ref_noisy)


if nps.norm2(rt_cam_ref[0]) != 0:
    print("First camera is assumed to sit at the origin, but it isn't there",
          file=sys.stderr)
    sys.exit()

# Add weight column. All weights are 1.0
observations_triangulated = nps.glue(observations,
                                     np.ones(observations.shape[:-1] + (1,),
                                             dtype = np.float32),
                                     axis = -1)

optimization_inputs = \
    dict( intrinsics            = nps.atleast_dims(m.intrinsics()[1], -2),
          rt_cam_ref            = rt_cam_ref[1:], # I made sure camera0 is at the origin

          observations_point_triangulated                        = observations_triangulated,
          indices_point_triangulated_camintrinsics_camextrinsics = indices_point_camintrinsics_camextrinsics,

          lensmodel                           = m.intrinsics()[0],
          imagersizes                         = nps.atleast_dims(m.imagersize(), -2),
          do_optimize_intrinsics_core         = False,
          do_optimize_intrinsics_distortions  = False,
          do_optimize_extrinsics              = True,
          do_optimize_frames                  = True,
          do_apply_outlier_rejection          = False,
          do_apply_regularization             = True,
          do_apply_regularization_unity_cam01 = True,
          verbose                             = False)

stats = mrcal.optimize(**optimization_inputs)

### We have a 5DOF solve. We rescale it to match the original
rt_cam_ref[:,3:] *= np.sqrt( nps.norm2(rt_cam_ref_true[-1,3:])/nps.norm2(rt_cam_ref[-1,3:]) )

if False:
    mrcal.show_geometry( nps.glue(rt_cam_ref, rt_cam_ref_true, axis=-2),
                         cameranames=[f"cam{i}-{what}" for what in ("solved","true") for i in range(Ncamera_poses)])

rt_err = mrcal.compose_rt( rt_cam_ref, rt_cam_ref_true, inverted1=True)

err_r_deg = nps.mag(rt_err[:,:3]) * 180./np.pi
err_t     = nps.mag(rt_err[:,3:])

testutils.confirm_equal(err_r_deg, 0,
                        worstcase = True,
                        eps = 2.,
                        msg = f"Recovered rotation")
testutils.confirm_equal(err_t, 0,
                        worstcase = True,
                        eps = 0.1,
                        msg = f"Recovered translation")


testutils.finish()
